#include <libtcod.hpp>
#include <stdio.h>
#include "build.h"
#include "map.h"
#include "vec2.h"
#include "atlas.h"
#include "buf.h"
#include "text.h"

class bu_vtx
{
public:
    bu_vtx() : pos(0, 0) { rad = 0; }
    bu_vtx(float ix, float iy) : pos(ix, iy) { rad = 0.0; }

    VEC2		pos;
    float		rad;
};

struct bu_edge
{
public:
    bu_edge() { vtx[0] = -1; vtx[1] = -1; }
    bu_edge(int a, int b) { vtx[0] = a; vtx[1] = b; }

    float		dist(VEC2 p, const PTRLIST<bu_vtx> &vtxlist)
    {
	VEC2		proj, linedir;
	proj = p;
	proj -= vtxlist(vtx[0]).pos;
	linedir = vtxlist(vtx[1]).pos;
	linedir -= vtxlist(vtx[0]).pos;

	float t = dot(proj, linedir) / linedir.quadrance();

	t = BOUND(t, 0, 1);
	float radius = lerp(t, vtxlist(vtx[0]).rad,
			       vtxlist(vtx[1]).rad);

	proj = vtxlist(vtx[0]).pos;
	proj += t * linedir;

	return ::dist(proj, p) - radius;
    }

    int			vtx[2];
};

static float
computeLineDist(VEC2 p, const PTRLIST<bu_edge> &edgelist, const PTRLIST<bu_vtx> &vtxlist)
{
    float	dist = 1e10f;
    for (int i = 0; i < edgelist.entries(); i++)
    {
	dist = MIN(dist, edgelist(i).dist(p, vtxlist));
    }
    return dist;
}

static void
createSplat(ROOMDATA<float> &data,
	    VEC2 splatcenter,
	    VEC2 splatsize,
	    int nlines,
	    float centerrad,
	    bool buildfingers, float fingerminrad, float fingermaxrad,
	    float noisesize, float noisedisplace)
{
    PTRLIST<bu_vtx>	 vtxlist;
    PTRLIST<bu_edge>	 edgelist;

    // First vertex is at 0,0
    vtxlist.append(bu_vtx(0, 0));

    for (int i = 0; i < nlines; i++)
    {
	// pick a random edge.
	int		startvtx = rand_choice(vtxlist.entries());
	bu_vtx		endpos;

	endpos.pos = vtxlist(startvtx).pos + VEC2::randUnit() * rand_frange(0.5, 1.5);
	int		endvtx = vtxlist.append(endpos);
	edgelist.append(bu_edge(startvtx, endvtx));
    }

    // Compute, first pass we compute our index count.
    for (int i = 0; i < edgelist.entries(); i++)
    {
	vtxlist(edgelist(i).vtx[0]).rad += 1.0;
	vtxlist(edgelist(i).vtx[1]).rad += 1.0;
    }

    // Now remap...
    for (int i = 0; i < vtxlist.entries(); i++)
    {
	vtxlist(i).rad -= 1;
	if (vtxlist(i).rad > 0.5)
	    vtxlist(i).rad = centerrad;
    }

    // Expand all fingers.
    if (buildfingers)
    {
	int		maxvtx = vtxlist.entries();
	for (int i = 0; i < maxvtx; i++)
	{
	    if (vtxlist(i).rad < 0.001)
	    {
		// pick a random edge.
		int		startvtx = i;
		bu_vtx	endpos;

		endpos.pos = vtxlist(startvtx).pos + VEC2::randUnit() * rand_frange(0.5, 1.5);
		endpos.rad = fingerminrad;
		int		endvtx = vtxlist.append(endpos);
		bu_vtx	startpos;
		startpos.pos = vtxlist(startvtx).pos;
		startpos.rad = fingermaxrad;
		startvtx = vtxlist.append(startpos);

		edgelist.append(bu_edge(startvtx, endvtx));
	    }
	}
    }

    VEC2		mincoord, maxcoord;

    mincoord = maxcoord = vtxlist(0).pos;
    for (int i = 1; i < vtxlist.entries(); i++)
    {
	mincoord = min(mincoord, vtxlist(i).pos);
	maxcoord = max(maxcoord, vtxlist(i).pos);
    }

    // Recenter & rescale to 10 .. 90
    float		size = (maxcoord - mincoord).maxcomp();
    VEC2		centeradjust = (maxcoord + mincoord) * 0.5;

    for (int i = 0; i < vtxlist.entries(); i++)
    {
	vtxlist(i).pos -= centeradjust;
	vtxlist(i).pos *= splatsize / size;
	vtxlist(i).pos += splatcenter;
    }

    // Allocate a noise map
    TCODRandom		r(rand_int());
    TCODNoise		noisex(2, &r), noisey(2, &r);

    // And compute a distance map.
    for (int y = 0; y < data.height(); y++)
    {
	for (int x = 0; x < data.width(); x++)
	{
	    VEC2		p((float)x, (float)y);

	    float		linepos[2];
	    linepos[0] = ((float)x)/noisesize;
	    linepos[1] = ((float)y)/noisesize;
	    float		dx, dy;
	    dx = noisedisplace * noisex.getTurbulenceSimplex(linepos, 3);
	    dy = noisedisplace * noisey.getTurbulenceSimplex(linepos, 3);
	    p.x() += dx;
	    p.y() += dy;

	    float		dist = computeLineDist(p, edgelist, vtxlist);
	    data(x, y) = MIN(data(x,y), dist);
	}
    }
}

ROOM_DEF *
BUILD::build(BUILD_NAME buildtype, int data, const char *name, const char *quest)
{
    ROOM_DEF_NC		*def = 0;

    
    switch (buildtype)
    {
	case BUILD_NONE:
	case NUM_BUILDS:
	    assert(0);
	    break;

	case BUILD_WORLD:
	    while (!def)
		def = buildWorld();
	    break;
	case BUILD_POI:
	    while (!def)
		def = buildPOI((POI_NAMES) data);
	    break;
    }


    if (quest)
	def->quest = strdup(quest);
    if (name)
	def->name = strdup(name);

    // Additional linking?
    switch (buildtype)
    {
	case BUILD_WORLD:
	    // World is self-wrapped.
	    def->west = def->north = def->south = def->east = def->name;
	    break;

	case BUILD_POI:
	    // POI always link to the left
	    if (glb_poidefs[data].backlink)
		def->west = strdup(glb_poidefs[data].backlink);
	    else
		def->west = strdup(glb_poidefs[data].name);

	    // Prizes always link to the right.
	    if (!def->east)
	    {
		BUF		prize;
		prize.sprintf("%s_prize", glb_poidefs[data].name);

		def->east = strdup(prize.buffer());
	    }
	    break;

	default:
	    break;
    }

    return (ROOM_DEF *) def;
}

ROOM_DEF_NC *
BUILD::buildPOI(POI_NAMES poi)
{
    ROOM_DEF_NC		*def, *lower;

    switch (poi)
    {
	case POI_MAZE:
	{
	    def = allocRoom(51, 51);
	    maze(def, SQUARE_FLOOR, SQUARE_WALL);
	    blockCorridors(def, 10, SQUARE_FLOOR, SQUARE_DOOR);

	    return def;
	}

	case POI_SECRET:
	{
	    def = allocRoom(21, 21);
	    maze(def, SQUARE_FLOOR, SQUARE_WALL);
	    blockCorridors(def, 10, SQUARE_FLOOR, SQUARE_SECRETDOOR);
	    blockCorridors(def, 10, SQUARE_FLOOR, SQUARE_DOOR);

	    return def;
	}

	case POI_DIZZY:
	{
	    def = allocRoom(21, 21);
	    maze(def, SQUARE_ROCKGROUND, SQUARE_WALL);
	    blockNodes(def, 20, SQUARE_ROCKGROUND, SQUARE_GROUND_SPINNER);

	    return def;
	}
	case POI_WATER:
	{
	    def = allocRoom(21, 21);
	    maze(def, SQUARE_GRASS, SQUARE_WATER);
	    blockCorridors(def, 20, SQUARE_GRASS, SQUARE_WATER);

	    return def;
	}
	case POI_LAVA:
	{
	    def = allocRoom(21, 21);
	    maze(def, SQUARE_GROUND, SQUARE_LAVA);
	    blockCorridors(def, 20, SQUARE_GROUND, SQUARE_LAVA);
	    int		ex, ey;
	    ex = def->size.x-1;
	    ey = randomSquareInColumn(def, SQUARE_LAVA, ex);
	    addMob(def, ex-1, ey, MOB_CAVETROLL);

	    return def;
	}
	case POI_FOREST:
	{
	    def = allocRoom(30, 25);
	    cavern(def, 50, SQUARE_FOREST, SQUARE_MOUNTAINS);
	    shrinkToFit(def, 1, 1, SQUARE_MOUNTAINS);

	    // Make our exits
	    int		px, py;
	    px = 0;
	    py = randomSquareInColumn(def, SQUARE_FOREST, px+1);
	    assert(py >= 0);
	    set(def, px, py, SQUARE_FOREST);
	    addLink(def, px, py, 100);

	    px = def->size.x-1;
	    py = randomSquareInColumn(def, SQUARE_FOREST, px-1);
	    assert(py >= 0);
	    set(def, px, py, SQUARE_FOREST);
	    addLink(def, px, py, 100);

	    // Spinify some forests.
	    replace(def, SQUARE_FOREST, SQUARE_FOREST_SPINNER, 20);

	    return def;
	}
	case POI_TELE:
	{
	    def = buildTeleMaze(4, 4, true);
	    return def;
	}
	case POI_TELESAME:
	{
	    def = buildTeleMaze(4, 4, false);
	    return def;
	}
	case POI_PORT:
	{
	    const int		leverlist[] =
	    {
		3, 1,
		5, 0,
		4, 1,
		5, 2,
		6, 3,		// HAVE YOU BEING PAYING ATTENTION?
		-1, -1
	    };
	    ROOM_DEF_NC		*ldef;
	    def = buildPort("port_gen", 3, 0);
	    ldef = def;
	    for (int gate = 0; leverlist[gate*2] >= 0; gate++)
	    {
		ROOM_DEF_NC		*ndef;

		if (leverlist[gate*2+2] < 0)
		    ldef->east = strdup("port_last");
		else
		    ldef->east = createUniqueName("port");
		ndef = buildPort(ldef->east, leverlist[gate*2], leverlist[gate*2+1]);
		ndef->west = strdup(ldef->name);

		registerDef(ndef);

		ldef = ndef;
	    }
	    ldef->east = strdup("port_prize");
	    return def;
	}
	case POI_HOLES:
	{
	    def = allocRoom(15, 15);
	    maze(def, SQUARE_FLOOR, SQUARE_FLOOR_HOLE);
	    replace(def, SQUARE_FLOOR_HOLE, SQUARE_FLOOR_HOLE_TRAP, 50);

	    def->down = strdup("hole_down");
	    lower = allocRoom(15, 15);
	    lower->name = strdup("hole_down");
	    fill(lower, SQUARE_WALL);
	    rect(lower, 1, 1, lower->size.x-2, lower->size.y-2,
			SQUARE_GROUND);

	    addMob(lower, 7, 7, MOB_SKELETON);
	    for (int i = 0; i < 5; i++)
	    {
		int		bx, by;
		if (randomSquare(lower, SQUARE_GROUND, bx, by))
		    addItem(lower, bx, by, ITEM_BONES);
	    }
	    registerDef(lower);
	    return def;
	}

	case POI_LADDER:
	{
	    def = buildLadderMaze(10, 10);
	    return def;
	}

	case POI_TROLL:
	{
	    def = buildTrollPrison();
	    return def;
	}

	case POI_PITFALL:
	{
	    // Looks a bit like a hole map, but has a ladder at
	    // the end and is impossible.
	    def = allocRoom(9, 9);
	    maze(def, SQUARE_FLOOR, SQUARE_FLOOR_HOLE);
	    replace(def, SQUARE_FLOOR_HOLE, SQUARE_FLOOR_HOLE_TRAP, 50);


	    def->down = strdup("pitfall_down");
	    lower = allocRoom(9, 9);
	    lower->name = strdup("pitfall_down");
	    lower->up = strdup("pitfall_gen");
	    fill(lower, SQUARE_WALL);
	    rect(lower, 1, 1, lower->size.x-2, lower->size.y-2,
			SQUARE_GROUND);

	    // find our exit.
	    int		ex, ey;
	    ex = 8;
	    ey = randomSquareInColumn(def, SQUARE_FLOOR, ex);

	    ex--;

	    addItem(def, ex, ey, ITEM_LADDERDOWN);
	    addItem(lower, ex, ey, ITEM_LADDERUP);

	    // Ensure ladder is blocked by hidden holes
	    int		dx, dy;
	    FORALL_8DIR(dx, dy)
	    {
		if (dx>0)
		    continue;
		if (get(def, ex+dx, ey+dy) == SQUARE_FLOOR)
		    set(def, ex+dx, ey+dy, SQUARE_FLOOR_HOLE_TRAP);
	    }

	    for (int i = 0; i < 3; i++)
	    {
		int		bx, by;
		if (randomSquare(lower, SQUARE_GROUND, bx, by))
		{
		    if (bx == ex && by == ey)
			continue;
		    addItem(lower, bx, by, ITEM_BONES);
		}
	    }

	    registerDef(lower);
	    return def;
	}
    }

    def = allocRoom(12, 12);

    fill(def, SQUARE_WALL);

    rect(def, 1, 1, 10, 10, SQUARE_FLOOR);
    set(def, 0, 5, SQUARE_FLOOR);
    addLink(def, 0, 5, 100);

    return def;
}

ROOM_DEF_NC *
BUILD::buildWorld()
{
    ROOM_DEF_NC		*def = allocRoom(120, 120);
    ROOM_DEF_NC		*underworld = 0;

    fill(def, SQUARE_WATER);

    ROOMDATA<float>	 mountainmap(def->size.x, def->size.y, 1000);

    createSplat(mountainmap, VEC2(60, 60), VEC2(45, 45),
			/*nlines=*/10,
			/*centerrad=*/2,
			/*buildfingers=*/true,
			/*fingerminmax=*/-10, -5,
			/*noisesize=*/10,
			/*noisedisplace=*/5
			);

    VEC2		islandcenter;
    float		displaceisland = rand_choice(2) ? 5.0f : 0.0f;
    islandcenter.y() = 35 - displaceisland;
    islandcenter.x() = rand_choice(2) ? 35  - displaceisland: 85 + displaceisland;
    float		centerrad = rand_choice(2) ? 5.0f : 0.5f;
    createSplat(mountainmap, islandcenter, VEC2(15, 15),
			/*nlines=*/5,
			/*centerrad=*/centerrad,
			/*buildfingers=*/false,
			/*fingerminmax=*/-10, -5,
			/*noisesize=*/10,
			/*noisedisplace=*/8
			);

    if (displaceisland < 0.5)
	displaceisland = 5;
    islandcenter.y() = 85 + displaceisland;
    islandcenter.x() = rand_choice(2) ? 35 - displaceisland : 85 + displaceisland;
    if (centerrad < 2.5)
	centerrad = 5;
    createSplat(mountainmap, islandcenter, VEC2(15, 15),
			/*nlines=*/5,
			/*centerrad=*/centerrad,
			/*buildfingers=*/false,
			/*fingerminmax=*/-10, -5,
			/*noisesize=*/10,
			/*noisedisplace=*/8
			);

    // And compute a distance map.
    for (int y = 0; y < def->size.y; y++)
    {
	for (int x = 0; x < def->size.x; x++)
	{
	    float		dist = mountainmap(x, y);
	    if (dist < 0)
		set(def, x, y, SQUARE_MOUNTAINS);
	    else if (dist < 5)
		set(def, x, y, SQUARE_HILLS);
	    else if (dist < 10)
		set(def, x, y, SQUARE_GRASS);
	}
    }

    // Find our tallest peak that will become a volcano.
    // By construction this should be the fat island.
    {
	int		peakx, peaky;
	mountainmap.findMin(peakx, peaky);

	// Verify we have sufficient mountains.
	rect(def, peakx-4, peaky-4, 9, 9, SQUARE_MOUNTAINS);

	// Now build our lava
	rect(def, peakx-2, peaky-2, 5, 5, SQUARE_LAVA);
	rect(def, peakx-3, peaky-1, 7, 3, SQUARE_LAVA);
	rect(def, peakx-1, peaky-3, 3, 7, SQUARE_LAVA);

	bool	didriver = false;
	for (int tries = 0; tries < 5; tries++)
	{
	    if (drawRiver(def, mountainmap, peakx, peaky, SQUARE_LAVA, SQUARE_WATER))
	    {
		didriver = true;
		break;
	    }
	}
	if (!didriver)
	{
	    delete def;
	    delete underworld;
	    return 0;
	}

	// And the rocky ground
	rect(def, peakx-1, peaky-1, 3, 3, SQUARE_ROCKGROUND);

	// And our hole!
	set(def, peakx, peaky, SQUARE_GROUND_HOLE);

	// And now a guardian!
	addMob(def, peakx-1, peaky, MOB_REDTRIDUDE);


	// Now build the underworld.
	underworld = allocRoom(def->size.x, def->size.y);

	underworld->globalmapflags = FLAG_DARK;

	cavern(underworld, 1,
		SQUARE_ROCKGROUND,
		SQUARE_SOLIDEARTH,
		peakx, peaky);

	int		prizex, prizey;
	prizex = peakx;
	prizey = peaky;
	findFarthest(underworld, SQUARE_ROCKGROUND, prizex, prizey);
	addItem(underworld, prizex, prizey, ITEM_BOOK_MALACHITETOWER);

	// Add some ghosts.
	for (int i = 0; i < 3; i++)
	{
	    int		mobx, moby;
	    if (randomSquare(underworld, SQUARE_ROCKGROUND, mobx, moby))
	    {
		addMob(underworld, mobx, moby, MOB_GHOST);
	    }
	}
    }

    ROOMDATA<int>	connectset(def->size.x, def->size.y, -1);
    PTRLIST<int>	connectcount;

    findConnected(def, false, false, connectset, connectcount);

    PTRLIST<int>	sortedcount;
    sortedcount = connectcount;
    sortedcount.stableSort();

    if (sortedcount.entries() < 2)
    {
	delete def;
	delete underworld;
	return 0;
    }
    int		maxcount = sortedcount(sortedcount.entries()-1);
    int		submaxcount = sortedcount(sortedcount.entries()-2);

    if (submaxcount < 200 || (maxcount == submaxcount))
    {
	delete def;
	delete underworld;
	return 0;
    }

    // Now find the biggest group.
    int		bigclass, secondclass;
    for (int i = 0; i < connectcount.entries(); i++)
    {
	if (connectcount(i) == maxcount)
	    bigclass = i;
	if (connectcount(i) = submaxcount)
	    secondclass = i;
    }

    // Verify our big classes touch lava.
    bool		touchlava = false;
    for (int y = 1; y < def->size.y-1; y++)
    {
	for (int x = 1; x < def->size.x-1; x++)
	{
	    int		dx, dy;
	    if (get(def, x, y) == SQUARE_LAVA)
	    {
		FORALL_4DIR(dx, dy)
		{
		    // We use 200 to flag a big area so we don't falsely
		    // think we are connected to a minor spurious piece of
		    // of noise.
		    int		neighbourset = connectset(x+dx, y+dy);
		    if (neighbourset >= 0 && connectcount(neighbourset) > 200)
		    {
			touchlava = true;
			break;
		    }
		}
	    }
	    if (touchlava)
		break;
	}
	if (touchlava)
	    break;
    }

    if (!touchlava)
    {
	delete def;
	delete underworld;
	return 0;
    }

    // Build some interesting biomes
#define NUM_BIOME 4
    const SQUARE_NAMES	biome[NUM_BIOME] = 
    { 
	SQUARE_FOREST,
	SQUARE_DESERT,
	SQUARE_FOREST,
	SQUARE_DESERT,
    };
    for (int curbiome = 0; curbiome < NUM_BIOME; curbiome++)
    {
	int		px, py;

	if (!connectset.randomNotEqual(-1, px, py))
	{
	    delete def;
	    delete underworld;
	    return 0;
	}
	ROOMDATA<float>	 biomemap(def->size.x, def->size.y, 1000);

	createSplat(biomemap, VEC2(px, py), VEC2(25, 25),
			/*nlines=*/8,
			/*centerrad=*/6,
			/*buildfingers=*/false,
			/*fingerminmax=*/-10, -5,
			/*noisesize=*/10,
			/*noisedisplace=*/12
			);

	// And apply the biome
	for (int y = 0; y < def->size.y; y++)
	{
	    for (int x = 0; x < def->size.x; x++)
	    {
		float		dist = biomemap(x, y);
		if (dist < 0)
		{
		    SQUARE_NAMES	curval;
		    curval = get(def, x, y);
		    if (curval == SQUARE_HILLS || curval == SQUARE_GRASS)
		    {
			set(def, x, y, biome[curbiome]);
		    }
		}
	    }
	}
    }
    // Assign our POI.
    POI_NAMES		poi;

    PTRLIST<PT2>	poilist;
    ROOMDATA<int>	poiconnect(def->size.x, def->size.y, -1);
    ROOMDATA<int>	hasitem(def->size.x, def->size.y, 0);

    poiconnect = connectset;
    FOREACH_POI(poi)
    {
	if (poi == POI_NONE)
	    continue;

	int		px, py;

	if (poi == POI_HOME)
	{
	    // This is always on the big island to give 
	    // some chances!
	    if (!poiconnect.randomEqual(bigclass, px, py))
	    {
		delete def;
		delete underworld;
		return 0;
	    }
	}
	else
	{
	    //int		poitries = 30;
	    //while (poitries --> 0)
	    {
		if (!poiconnect.randomNotEqual(-1, px, py))
		{
		    delete def;
		    delete underworld;
		    return 0;
		}
		// Verify we have over 3 cells
		// You need to have room to step off the place
		// to be fair.
		// But who really cares about fair?
		// You have the recall spell!
	    }

	    poilist.append(point(px, py, poi));
	}

	addItem(def, px, py, ITEM_HUT);
	hasitem(px, py) = ITEM_HUT;
	addLink(def, px, py, poi);

	// Prevent POI from being too close.
	poiconnect.clearIfEqual(connectset(px, py), -1,
			    px, py, 10, 10);
    }

    // Build prizes!
    PTRLIST<SPELL_NAMES>		spellbooks;
    PTRLIST<POI_NAMES>		diskbooks;
    PTRLIST<int>			poi2poiidx;

    int		prizedeptries = 100;
    while (prizedeptries --> 0)
    {
	poi2poiidx.resize(NUM_POIS);
	poi2poiidx.constant(-1);
	for (int i = 0; i < poilist.entries(); i++)
	    poi2poiidx(poilist(i).data) = i;

	spellbooks.resize(poilist.entries());
	spellbooks.constant(SPELL_NONE);
	diskbooks.resize(poilist.entries());
	diskbooks.constant(POI_NONE);

	PTRLIST<int>		diskpermute;
	buildTruePermutation(diskpermute, diskbooks.entries());
	for (int i = 0; i < diskbooks.entries(); i++)
	{
	    // The ith POI will give the secret to diskpermute(i)th POI
	    diskbooks(i) = (POI_NAMES) poilist(diskpermute(i)).data;
	}
	// Validate you can get from home to all POI we care about.
	// We require reachability from vanilla rooms.
	PTRLIST<bool>		reachable;
	reachable.resize(connectcount.entries());
	reachable.constant(false);
	reachable(bigclass) = true;
	bool		changed = true;
	while (changed)
	{
	    changed = false;
	    for (int i = 0; i < diskbooks.entries(); i++)
	    {
		if (!reachable(connectset(poilist(i).x, poilist(i).y)))
		    continue;

		// If any spell requirements, ignore
		if (*glb_poidefs[poilist(i).data].spelldep)
		    continue;

		int	destpoiidx = diskpermute(i);
		int	destset = connectset(poilist(destpoiidx).x, poilist(destpoiidx).y);
		// See if already found
		if (reachable(destset))
		    continue;

		changed = true;
		reachable(destset) = true;
	    }
	}
	// Stopped changing so we searched the graph
	// Ensure we got everything!
	bool	failed = false;
	for (int i = 0; i < diskbooks.entries(); i++)
	{
	    if (!reachable(connectset(poilist(i).x, poilist(i).y)))
		failed = true;
	}

	if (failed)
	{
	    // Try again!
	    continue;
	}

	// Now assign spells.
	// We can't give out a spell that is our own requirement
	// But when we give out a spell, any POI with that spell
	// as a requirement gains our own requirements as well!
	// In both these cases we have a dependency loop, we can
	// break it by reassigning our own prize 

	// Determine first generation illegality.

	PTRLIST<bool>		legalspells[NUM_POIS];
	FOREACH_POI(poi)
	{
	    legalspells[poi].resize(NUM_SPELLS);
	    legalspells[poi].constant(false);

	    if (poi == POI_NONE)
		continue;
	    if (poi == POI_HOME)
		continue;

	    legalspells[poi].constant(true);

	    legalspells[poi](SPELL_RECALL) = false;

	    // Prohibit our own parents.
	    for (const char *sp = glb_poidefs[poi].spelldep; *sp; sp++)
	    {
		legalspells[poi](*sp) = false;
	    }
	}


	// Now assign all spells.
	PTRLIST<SPELL_NAMES>	prizespells;
	SPELL_NAMES	spell;
	FOREACH_SPELL(spell)
	{
	    if (spell == SPELL_NONE)
		continue;
	    if (spell == SPELL_RECALL)
		continue;		// Handed out by home!
	    prizespells.append(spell);
	}
	prizespells.shuffle();

	for (int i = 0; i < prizespells.entries(); i++)
	{
	    spell = prizespells(i);

	    // Find a random legal room to assign to!
	    POI_NAMES		spellroom = POI_NONE;
	    int			choice = 0;
	    FOREACH_POI(poi)
	    {
		if (!legalspells[poi](spell))
		    continue;
		if (spellbooks(poi2poiidx(poi)) != SPELL_NONE)
		    continue;

		choice++;
		if (!rand_choice(choice))
		    spellroom = poi;
	    }

	    if (spellroom == POI_NONE)
	    {
		failed = true;
		break;
	    }

	    spellbooks(poi2poiidx(spellroom)) = spell;

	    // Now propagate the dependency!
	    // Any room which cannot legally have spell
	    // must now also not have any spells that are
	    // illegal to myself.
	    // We do not need to look at the spelldep as we've
	    // collapsed this into the legalspells list.
	    FOREACH_POI(poi)
	    {
		if (poi == spellroom)
		    continue;

		if (!legalspells[poi](spell))
		{
		    // THis room cannot have spell, therefore
		    // it cannot have anything spellroom cares about
		    SPELL_NAMES	s;
		    FOREACH_SPELL(s)
		    {
			legalspells[poi](s) &= legalspells[spellroom](s);
		    }
		}
	    }
	}

	if (failed)
	    continue;

#if 0
	for (int i = 0; i < poilist.entries(); i++)
	{
	    printf("POI %s gives Spell %s\n",
		    glb_poidefs[poilist(i).data].name,
		    glb_spelldefs[spellbooks(i)].words);
	}
#endif
	break;
    }

    if (prizedeptries <= 0)
    {
	delete def;
	delete underworld;
	return 0;
    }

    // Now that we can't fail, hand out the prizes!
    for (int i = 0; i < poilist.entries(); i++)
    {
	POI_NAMES	poi = (POI_NAMES) poilist(i).data;

	ROOM_DEF_NC	*prize = allocRoom(5, 5);
	BUF		 buf;

	fill(prize, (SQUARE_NAMES) glb_poidefs[poi].wall);
	rect(prize, 1, 1, 3, 3, (SQUARE_NAMES) glb_poidefs[poi].floor);
	set(prize, 0, 2, (SQUARE_NAMES) glb_poidefs[poi].floor);
	addLink(prize, 0, 2, 100);

	if (glb_poidefs[poi].prizelink)
	    prize->west = strdup(glb_poidefs[poi].prizelink);
	else
	{
	    buf.sprintf("%s_gen", glb_poidefs[poi].name);
	    prize->west = strdup(buf.buffer());
	}
	buf.sprintf("%s_prize", glb_poidefs[poi].name);
	prize->name = strdup(buf.buffer());

	ITEM_NAMES		founditem = ITEM_NONE;
	ITEM_NAMES		item;
	if (spellbooks(i) != SPELL_NONE)
	{
	    founditem = ITEM_NONE;
	    FOREACH_ITEM(item)
	    {
		if (glb_itemdefs[item].spell == spellbooks(i))
		    founditem = item;
	    }
	    assert(founditem != ITEM_NONE);
	    addItem(prize, 3, 3, founditem);
	}

	founditem = ITEM_NONE;
	FOREACH_ITEM(item)
	{
	    if (glb_itemdefs[item].poi == diskbooks(i))
		founditem = item;
	}
	assert(founditem != ITEM_NONE);
	addItem(prize, 3, 1, founditem);

	registerDef(prize);
    }

    // Count all of our sign posts
    int		numsignpost = 0;
    while (1)
    {
	BUF		post;
	post.sprintf("%d", numsignpost);
	if (!text_hasentry("sign", post))
	    break;
	numsignpost++;
    }

    for (int i = 0; i < numsignpost; i++)
    {
	int		signtries = 10;
	int		px, py;

	// Not the end of the world if this fails
	while (signtries --> 0)
	{
	    if (!poiconnect.randomNotEqual(-1, px, py))
	    {
		continue;
	    }
	    // ENsure no item already here
	    if (hasitem(px, py) == ITEM_NONE)
		break;
	}

	if (signtries > 0)
	{
	    assert(hasitem(px, py) == ITEM_NONE);

	    hasitem(px, py) = ITEM_BOULDERSIGN;
	    addItem(def, px, py, ITEM_BOULDERSIGN);
	    // Yes, you can add 0.
	    addLink(def, px, py, i);
	}
    }

    replace(def, SQUARE_HILLS, SQUARE_FOREST, 5);
    replace(def, SQUARE_GRASS, SQUARE_FOREST, 2);
    replace(def, SQUARE_DESERT, SQUARE_CACTUS, 1);

    // Link in the underworld!
    underworld->name = strdup("underworld");
    def->down = strdup(underworld->name);
    // Note no way up!
    registerDef(underworld);

    return def;
}

void
BUILD::set(ROOM_DEF_NC *def, int x, int y, SQUARE_NAMES square)
{
    assert(x >= 0 && y >= 0);
    assert(x < def->size.x && y < def->size.y);

    def->squarelist[x + y * def->size.x] = square;
}

SQUARE_NAMES
BUILD::get(const ROOM_DEF_NC *def, int x, int y)
{
    if (x < 0 || y < 0 || x >= def->size.x || y >= def->size.y)
    {
	return SQUARE_NONE;
    }

    return def->squarelist[x + y * def->size.x];
}

void
BUILD::fill(ROOM_DEF_NC *def, SQUARE_NAMES square)
{
    rect(def, 0, 0, def->size.x, def->size.y, square);
}

void
BUILD::rect(ROOM_DEF_NC *def, int sx, int sy, int w, int h, SQUARE_NAMES square)
{
    for (int y = 0; y < h; y++)
    {
	if (y + sy < 0 || y + sy >= def->size.y)
	    continue;
	for (int x = 0; x < w; x++)
	{
	    if (x + sx < 0 || x + sx >= def->size.x)
		continue;
	    set(def, x+sx, y+sy, square);
	}
    }
}

ROOM_DEF_NC *
BUILD::allocRoom(int w, int h)
{
    ROOM_DEF_NC		*def = new ROOM_DEF_NC;

    def->size.x = w;
    def->size.y = h;
    def->squarelist = new SQUARE_NAMES[w * h];

    def->squareflaglist = allocPtList();
    def->itemlist = allocPtList();
    def->moblist = allocPtList();
    def->itemtypelist = allocPtList();
    def->moblevellist = allocPtList();
    def->intrinsiclist = allocPtList();
    def->signpostlist = allocPtList();
    def->linklist = allocPtList(-1);

    def->minlevel = def->maxlevel = 0;
    def->rarity = 100;
    def->globalmapflags = 0;
    def->onlyonce = 0;
    def->allowstairs = 0;
    def->allowgeneration = 0;
    def->allowdig = 1;
    def->allowitems = 0;
    def->allowtel = 1;
    def->x = -1;
    def->y = -1;

    def->quest = 0;
    def->north = def->east = def->south = def->west = def->up = def->down = 0;
    def->name = 0;
    
    fill(def, SQUARE_NONE);

    return def;
}

PT2 *
BUILD::allocPtList(int lastdata)
{
    PT2		*list = new PT2[1];
    list->x = list->y = -1;
    list->data = lastdata;

    return list;
}

PT2 *
BUILD::appendPtList(PT2 *oldlist, PT2 newpt)
{
    int		 len;
    PT2		*list;

    for (len = 0; oldlist[len].data > 0; len++);
    len++;		// include null


    list = new PT2[len+1];
    memcpy(list, oldlist, len * sizeof(PT2));
    delete [] oldlist;
    // Copy forward null
    list[len] = list[len-1];
    // And insert our new point
    list[len-1] = newpt;

    return list;
}

static int
findClass(ROOMDATA<int> &fclass, int x, int y)
{
    int		c = fclass(x, y);
    if (c < 0)
	return c;

    int		linidx = x + y * fclass.width();
    if (c == linidx)
	return c;

    int		nx, ny;
    nx = c % fclass.width();
    ny = c / fclass.width();

    c = findClass(fclass, nx, ny);
    fclass(x, y) = c;

    return c;
}

void
unionSquares(ROOMDATA<int> &fclass, int x1, int y1, int x2, int y2)
{
    int		c1 = findClass(fclass, x1, y1);
    int		c2 = findClass(fclass, x2, y2);
    // We union the new tree with the old tree, 
    // this then preserves the rank order.
    if (c1 == c2)
	return;
    if (c1 < c2)
    {
	int		nx, ny;
	nx = c2 % fclass.width();
	ny = c2 / fclass.width();
	fclass(nx, ny) = c1;
    }
    else
    {
	int		nx, ny;
	nx = c1 % fclass.width();
	ny = c1 / fclass.width();
	fclass(nx, ny) = c2;
    }
}

void
BUILD::findConnected(ROOM_DEF_NC *def, bool allowwater, bool allowfire,
		ROOMDATA<int> &connect, PTRLIST<int> &count)
{
    int			w = def->size.x, h = def->size.y;
    ROOMDATA<int>	fclass(w, h, -1);

    for (int y = 0; y < h; y++)
    {
	for (int x = 0; x < w; x++)
	{
	    // Illegal tiles get -1
	    SQUARE_NAMES	square = get(def, x, y);
	    if (glb_squaredefs[square].connectable)
	    {
		// See if we match the water and fire.
		if ( (!glb_squaredefs[square].connectable_fire || allowfire) &&
		     (!glb_squaredefs[square].connectable_water || allowwater) )
		{
		    fclass(x, y) = x + y * w;
		}
	    }
	}
    }

    // Two way connectivity.
    for (int y = 0; y < h; y++)
    {
	for (int x = 0; x < w; x++)
	{
	    if (findClass(fclass, x, y) < 0)
		continue;

	    if (x > 0 && findClass(fclass, x-1, y) >= 0)
	    {
		// Connect back.
		unionSquares(fclass, x-1, y, x, y);
	    }
	    if (y > 0 && findClass(fclass, x, y-1) >= 0)
	    {
		unionSquares(fclass, x, y-1, x, y);
	    }
	}
    }

    // Create our lookup table to remap our classes.
    ROOMDATA<int>	lut(w, h, -1);
    int			nclass = 0;
    for (int y = 0; y < h; y++)
    {
	for (int x = 0; x < w; x++)
	{
	    int c = findClass(fclass, x, y);

	    if (c >= 0)
	    {
		// See if this class has a lut entry yet.
		int nx = c % w;
		int ny = c / w;
		if (lut(nx, ny) == -1)
		{
		    lut(nx, ny) = nclass++;
		}
	    }
	}
    }

    // Now remap all the classes and count!
    count.resize(0);
    count.resize(nclass);

    for (int y = 0; y < h; y++)
    {
	for (int x = 0; x < w; x++)
	{
	    int c = findClass(fclass, x, y);

	    if (c >= 0)
	    {
		int nx = c % w;
		int ny = c / w;

		int	value = lut(nx, ny);
		count(value)++;
		connect(x, y) = value;
	    }
	    else
		connect(x, y) = -1;
	}
    }
}

void
BUILD::addLink(ROOM_DEF_NC *def, int x, int y, int link)
{
    PT2		pt;
    pt.x = x;
    pt.y = y;
    pt.data = link;

    def->linklist = appendPtList(def->linklist, pt);
}

void
BUILD::addItem(ROOM_DEF_NC *def, int x, int y, ITEM_NAMES item)
{
    PT2		pt;
    pt.x = x;
    pt.y = y;
    pt.data = item;

    def->itemlist = appendPtList(def->itemlist, pt);
}

void
BUILD::addMob(ROOM_DEF_NC *def, int x, int y, MOB_NAMES mob)
{
    PT2		pt;
    pt.x = x;
    pt.y = y;
    pt.data = mob;

    def->moblist = appendPtList(def->moblist, pt);
}

void
BUILD::replace(ROOM_DEF_NC *def,
		SQUARE_NAMES oldsquare,
		SQUARE_NAMES newsquare,
		int chance)
{
    rectReplace(def, 0, 0, def->size.x, def->size.y,
		    oldsquare, newsquare, chance);
}

void
BUILD::rectReplace(ROOM_DEF_NC *def,
		int sx, int sy, int w, int h,
		SQUARE_NAMES oldsquare,
		SQUARE_NAMES newsquare,
		int chance)
{
    for (int y = 0; y < h; y++)
    {
	if (y + sy < 0 || y + sy >= def->size.y)
	    continue;
	for (int x = 0; x < w; x++)
	{
	    if (x + sx < 0 || x + sx >= def->size.x)
		continue;
	    if (get(def, x+sx, y+sy) == oldsquare)
	    {
		if (rand_chance(chance))
		    set(def, x+sx, y+sy, newsquare);
	    }
	}
    }
}

bool
BUILD::findRandomMazePassage(ROOM_DEF_NC *def, SQUARE_NAMES floor, int &rnx, int &rny, int &rdx, int &rdy)
{
    int			nvalid = 0;

    for (int y = 1; y < def->size.y; y += 2)
    {
	for (int x = 1; x < def->size.x; x += 2)
	{
	    if (get(def, x, y) == floor)
		continue;

	    int		dx, dy;

	    FORALL_4DIR(dx, dy)
	    {
		if (get(def, x + dx*2, y + dy*2) == floor)
		{
		    nvalid++;
		    if (!rand_choice(nvalid))
		    {
			rnx = x;		rny = y;
			rdx = dx;		rdy = dy;
		    }
		}
	    }
	}
    }

    return nvalid > 0;
}

int
BUILD::numExits(ROOM_DEF_NC *def, SQUARE_NAMES floor, int x, int y) 
{
    int		nexit = 0;
    int		dx, dy;

    FORALL_4DIR(dx, dy)
    {
	if (get(def, x+dx, y+dy) == floor)
	    nexit++;
    }
    return nexit;
}


void
BUILD::tryBlockCorridor(ROOM_DEF_NC *def, int x, int y, int chance, SQUARE_NAMES floor, SQUARE_NAMES blocker)
{
    if (get(def, x, y) == floor)
    {
	// Could be a corridor..
	// Either vertical are free or horizontal are free.
	bool		floordir[4];
	int		dx, dy;
	FORALL_4DIR(dx, dy)
	{
	    floordir[lcl_angle] = (get(def, x+dx, y+dy) == floor);
	}

	if ( (floordir[0] && floordir[2] && !floordir[1] && !floordir[3]) ||
	     (!floordir[0] && !floordir[2] && floordir[1] && floordir[3]))
	{
	    // Corridor!
	    if (rand_chance(chance))
		set(def, x, y, blocker);
	}
    }
}

void
BUILD::cavern(ROOM_DEF_NC *def, int percentfloor, SQUARE_NAMES floor, SQUARE_NAMES wall)
{
    set(def, def->size.x/2, def->size.y/2, floor);
    cavern(def, percentfloor, floor, wall, 
	    def->size.x/2, def->size.y/2);
}

void
BUILD::cavern(ROOM_DEF_NC *def, int percentfloor, SQUARE_NAMES floor, SQUARE_NAMES wall, int centerx, int centery)
{
    int		totaltoeat = (def->size.x-2) * (def->size.y-2) * percentfloor / 100;
    int		eaten = 0;

    fill(def, wall);
    PTRLIST<PT2>	starts;

    set(def, centerx, centery, floor);
    eaten++;
    starts.append(point(centerx, centery));

    while (eaten < totaltoeat)
    {
	int		ex, ey;
	int		choice;

	choice = rand_choice(starts.entries());
	ex = starts(choice).x;
	ey = starts(choice).y;

	int		len = rand_range(10, 20);
	int		dir = rand_range(0, 3);
	while (len--)
	{
	    int		dx, dy;
	    rand_getdirection(dir, dx, dy);
	    ex += dx;
	    ey += dy;
	    if (ex < 1 || ex >= def->size.x-1)
		break;
	    if (ey < 1 || ey >= def->size.y-1)
		break;

	    if (get(def, ex, ey) == wall)
	    {
		set(def, ex, ey, floor);
		starts.append(point(ex, ey));
		eaten++;
	    }

	    // Change direction.
	    // Do not backtrack!
	    dir += rand_choice(3) - 1;
	    dir &= 3;
	}
    }

}

void
BUILD::blockNodes(ROOM_DEF_NC *def, int chance, SQUARE_NAMES floor, SQUARE_NAMES blocker)
{
    // Assuming a maze, run over all nodes and block 
    // sometimes.
    for (int y = 1; y < def->size.y; y += 2)
    {
	for (int x = 1; x < def->size.x; x += 2)
	{
	    if (get(def, x, y) == floor)
	    {
		if (rand_chance(chance))
		    set(def, x, y, blocker);
	    }
	}
    }
}

void
BUILD::blockCorridors(ROOM_DEF_NC *def, int chance, SQUARE_NAMES floor, SQUARE_NAMES blocker)
{
    // Assuming a maze, run over all corridors and block straight
    // lines... sometimes.
    for (int y = 2; y < def->size.y; y += 2)
    {
	for (int x = 2; x < def->size.x; x += 2)
	{
	    tryBlockCorridor(def, x-1, y, chance, floor, blocker);
	    tryBlockCorridor(def, x, y-1, chance, floor, blocker);
	}
    }
    // SPECIAL CASE: Exit is always blocked!
    for (int y = 0; y < def->size.y; y++)
    {
	int x = def->size.x-1;
	if (get(def, x, y) == floor)
	    set(def, x, y, blocker);
    }
}

void
BUILD::maze(ROOM_DEF_NC *def, SQUARE_NAMES floor, SQUARE_NAMES wall)
{
    fill(def, SQUARE_WALL);
    rect(def, 1, 1, def->size.x-2, def->size.y-2, wall);

    int		sx, sy;

    sx = 0;
    sy = rand_range(0, def->size.y/2 - 1);
    sy = sy * 2 + 1;

    set(def, sx, sy, floor);
    addLink(def, sx, sy, 100);

    set(def, sx+1, sy, floor);

    sx = def->size.x-1;
    sy = rand_range(0, def->size.y/2 - 1);
    sy = sy * 2 + 1;
    set(def, sx, sy, floor);
    addLink(def, sx, sy, 100);
    set(def, sx-1, sy, floor);

    // Start with disconnected cells.
    int			nx, ny, dx, dy;
    while (findRandomMazePassage(def, floor, nx, ny, dx, dy))
    {
	set(def, nx, ny, floor);
	set(def, nx+dx, ny+dy, floor);
    }

    // Now open all the dead ends.
    for (int y = 1; y < def->size.y; y += 2)
    {
	for (int x = 1; x < def->size.x; x += 2)
	{
	    if (numExits(def, floor, x, y) == 1)
	    {
		int		nescape = 0, edx, edy;
		FORALL_4DIR(dx, dy)
		{
		    if (get(def, x + dx*2, y + dy*2) == floor)
		    {
			nescape++;
			if (!rand_choice(nescape))
			{
			    edx = dx;
			    edy = dy;
			}
		    }
		}
		if (nescape)
		{
		    set(def, x+edx, y+edy, floor);
		}
	    }
	}
    }

}

void
BUILD::registerDef(ROOM_DEF_NC *def)
{
    MAP		*map = new MAP(* (ROOM_DEF *)def);
    glbAtlas()->registerMap(map);
}

int
BUILD::countSquare(ROOM_DEF_NC *def, SQUARE_NAMES square)
{
    int		count = 0;
    for (int y = 0; y < def->size.y; y++)
	for (int x = 0; x < def->size.x; x++)
	{
	    if (get(def, x, y) == square)
	    {
		count++;
	    }
	}
    return count;
}

bool
BUILD::nthSquare(ROOM_DEF_NC *def, int nth, SQUARE_NAMES square, int &x, int &y)
{
    for (y = 0; y < def->size.y; y++)
	for (x = 0; x < def->size.x; x++)
	{
	    if (get(def, x, y) == square)
	    {
		if (!nth)
		    return true;
		nth--;
	    }
	}
    return false;
}

bool
BUILD::randomSquare(ROOM_DEF_NC *def, SQUARE_NAMES square, int &x, int &y)
{
    int		num = countSquare(def, square);

    x = y = -1;
    if (!num)
	return false;

    return nthSquare(def, rand_choice(num), square, x, y);
}

int
BUILD::randomSquareInColumn(ROOM_DEF_NC *def, SQUARE_NAMES square, int col)
{
    int		result = -1;
    int		total = 0;

    for (int y = 0; y < def->size.y; y++)
    {
	if (get(def, col, y) == square)
	{
	    total++;
	    if (!rand_choice(total))
		result = y;
	}
    }
    return result;
}

void
BUILD::shrinkToFit(ROOM_DEF_NC *def, int xpad, int ypad, SQUARE_NAMES wall)
{
    int		left, right;
    int		top, bot;
    int		x, y;

    for (left = 0; left < def->size.x; left++)
    {
	for (y = 0; y < def->size.y; y++)
	{
	    if (get(def, left, y) != wall)
		break;
	}
	if (y != def->size.y)
	    break;
    }

    for (right = def->size.x; right --> 0; )
    {
	for (y = 0; y < def->size.y; y++)
	{
	    if (get(def, right, y) != wall)
		break;
	}
	if (y != def->size.y)
	    break;
    }

    for (top = 0; top < def->size.y; top++)
    {
	for (x = 0; x < def->size.x; x++)
	{
	    if (get(def, x, top) != wall)
		break;
	}
	if (x != def->size.x)
	    break;
    }

    for (bot = def->size.y; bot --> 0; )
    {
	for (x = 0; x < def->size.x; x++)
	{
	    if (get(def, x, bot) != wall)
		break;
	}
	if (x != def->size.x)
	    break;
    }

    int		ow = def->size.x;
    int		oh = def->size.y;
    SQUARE_NAMES	*os = def->squarelist;

    def->size.x = (right - left) + 1 + 2*xpad;
    def->size.y = (bot - top) + 1 + 2*ypad;

    def->squarelist = new SQUARE_NAMES[def->size.x * def->size.y];

    fill(def, wall);

    // Now copy in center!
    for (y = 0; y < bot-top+1; y++)
    {
	for (x = 0; x < right-left+1; x++)
	{
	    set(def, x+xpad, y+ypad, 
		    os[ (x + left) + (y + top) * ow ]);
	}
    }

    delete [] os;
}

ROOM_DEF_NC *
BUILD::buildPort(const char *name, int nport, int nhidden)
{
    // We build a puzzle with nporticullis in a row.
    // We hide it with a separate room accessed via a trick.
    ROOM_DEF_NC		*def = allocRoom(1 + nport + nport + 1, 20);

    def->name = strdup(name);
    
    fill(def, SQUARE_WALL);

    // Entrance.
    set(def, 0, 10, SQUARE_FLOOR);
    addLink(def, 0, 10, 100);

    // A simple porticullis room.
    // South wall has levers nport levers
    rect(def, 1, 9, nport, 3, SQUARE_FLOOR);
    for (int i = nhidden; i < nport; i++)
    {
	set(def, 1 + i, 12, rand_choice(2) ? SQUARE_WALL_LEVER_UP : SQUARE_WALL_LEVER_DOWN);
	addLink(def, 1+i, 12, i+1);
    }

    if (nhidden)
    {
	// Build the ancillary room, protected by a porticullis!
	rect(def, nport - nhidden-1, 6, nhidden+1, 2, SQUARE_FLOOR);
	set(def,  nport-1, 8, SQUARE_PORTICULLIS_DOWN);
	// And draw the hidden levers.
	for (int i = 0; i < nhidden; i++)
	{
	    set(def, nport - nhidden + i, 5, rand_choice(2) ? SQUARE_WALL_LEVER_UP : SQUARE_WALL_LEVER_DOWN);
	    addLink(def, nport - nhidden + i, 5, i+1);
	}

	// Now decide the trick!
	switch (rand_choice(NUM_PORTTRICKS))
	{
	    case PORTTRICK_LEVER:
	    {
		addLink(def, nport-1, 8, nport+2);
		set(def, nport-2, 8, rand_choice(2) ? SQUARE_WALL_LEVER_UP : SQUARE_WALL_LEVER_DOWN);
		addLink(def, nport-2, 8, nport+2);
		break;
	    }

	    case PORTTRICK_TELE:
	    {
		addLink(def, 1, 9, nport+2);
		set(def, 1, 9, SQUARE_FLOOR_TELE);

		addLink(def, nport - nhidden-1, 7, nport+2);
		set(def, nport - nhidden-1, 7, SQUARE_FLOOR_TELE);
		break;
	    }

	    case PORTTRICK_LADDER:
	    {
		addItem(def, 1, 9, ITEM_LADDERDOWN);
		addItem(def, nport - nhidden-1, 7, ITEM_LADDERDOWN);

		ROOM_DEF_NC	*lower;
		lower = allocRoom(def->size.x, def->size.y);
		fill(lower, SQUARE_WALL);
		rect(lower, 1, 7, (nport-nhidden-1) - 1 + 1, 3, SQUARE_ROCKGROUND);

		addItem(lower, 1, 9, ITEM_LADDERUP);
		addItem(lower, nport - nhidden-1, 7, ITEM_LADDERUP);
		lower->name = createUniqueName("port_down");

		lower->up = strdup(name);
		def->down = strdup(lower->name);
		registerDef(lower);
		break;
	    }
	    default:
		break;
	}
    }

    // Porticullis immediately follow

    // Determine our solution
    int		soln;
    int		mask = (1 << (nport)) - 1;

    PTRLIST<int>	leverlist, doorlist;
    for (int i = 0; i < nport; i++)
    {
	leverlist.append(i);
	doorlist.append(i);
    }

    // For each lever, which doors it flips.
    PTRLIST<int>	lever2door;
    lever2door.resize(nport);
    lever2door.constant(0);

    // First, ensure every door is controlled by a lever.
    doorlist.shuffle();
    for (int i = 0; i < nport; i++)
	lever2door(i) |= (1 << doorlist(i));

    // Progressively select a fewer number of doors/levers to get
    // extra applications
    for (int j = 1; j < nport-1; j++)
    {
	doorlist.shuffle();
	leverlist.shuffle();
	for (int i = 0; i < nport - j; i++)
	{
	    lever2door(leverlist(i)) |= (1 << doorlist(i));
	}
    }

    // Determine effect of doors.
    int		solneffect = 0;

    // Ensure we aren't 0 or mask
    while (1)
    {
	soln = rand_choice(mask-1);
	soln++;

	// It is tempting to whiten this further, but that then
	// just leads to degenerate play trying to hack the whiten
	// restrictions.
	// However, one that is fair is to say it is never a single
	// lever to toggle - this kills a common procedure of just
	// toggling each lever once to see what happens.
	// This does not prohibit there being another solution
	// which is a single gate!
	if (! ((soln - 1) & soln))
	    continue;

	// Now see what the solution does!
	solneffect = 0;
	for (int i = 0; i < nport; i++)
	{
	    if (soln & (1 << i))
		solneffect ^= lever2door(i);
	}

	// If the effect is zero, we have an already open puzzle,
	// which is boring!!!  So we continue!
	if (!solneffect)
	    continue;

	break;
    }

    // Initial states must be the inverse of the solution effect.
    for (int i = 0; i < nport; i++)
    {
	int		doormask = 1 << i;
	if (solneffect & doormask)
	    set(def, 1 + i + nport, 11, SQUARE_PORTICULLIS_DOWN);
	else
	    set(def, 1 + i + nport, 11, SQUARE_PORTICULLIS_UP);
	// Apply proper links
	for (int j = 0; j < nport; j++)
	{
	    if (lever2door(j) & doormask)
	    {
		addLink(def, 1+i + nport, 11, j+1);
	    }
	}
    }

#ifdef STILL_TESTING
    // I'll have you know I have solved it properly!
    printf("Solution for %s (%d): ", name, soln);
    for (int i = 0; i < nport; i++)
    {
	if (soln & (1 << i))
	    printf("X");
	else
	    printf("O");
    }
    printf("\n");
#endif

    // Exit
    set(def, 1 + nport + nport, 11, SQUARE_FLOOR);
    addLink(def, 1 + nport + nport, 11, 100);

#ifdef STILL_TESTING
    // Validate our links.
    {
	MAP		*map;
	map = new MAP(*(ROOM_DEF *)def);

	int failed = map->validateLinks(false);
	if (failed)
	{
	    printf("Failed building %s, %d failures\n",
		    name, failed);
	    assert(!"Failed to validate!");
	}
	delete map;
    }
#endif

    return def;
}

char *
BUILD::createUniqueName(const char *prefix)
{
    static int suffix = 0;

    BUF		buf;
    buf.sprintf("%s_%d", prefix, suffix);
    suffix++;

    return strdup(buf.buffer());
}

ROOM_DEF_NC *
BUILD::buildTeleMaze(int maze_width, int maze_height, bool indicators)
{
    ROOM_DEF_NC		*def = allocRoom( 8 * maze_width, 8 * maze_height );

    fill(def, SQUARE_WALL);

    for (int y = 0; y < maze_height; y++)
    {
	for (int x = 0; x < maze_width; x++)
	{
	    int		roomnum = x + y * maze_height;
	    rect(def, x * 8 + 2, y * 8 + 2, 4, 4, SQUARE_FLOOR);
	    set(def, x * 8 + 2, y * 8 + 2, SQUARE_FLOOR_TELE);
	    set(def, x * 8 + 5, y * 8 + 2, SQUARE_FLOOR_TELE);
	    set(def, x * 8 + 2, y * 8 + 5, SQUARE_FLOOR_TELE);
	    set(def, x * 8 + 5, y * 8 + 5, SQUARE_FLOOR_TELE);
	    if (indicators)
	    {
		if (roomnum & 1)
		    rect(def, x * 8 + 1, y * 8 + 3, 1, 2, SQUARE_FLOOR);
		if (roomnum & 2)
		    rect(def, x * 8 + 3, y * 8 + 1, 2, 1, SQUARE_FLOOR);
		if (roomnum & 4)
		    rect(def, x * 8 + 6, y * 8 + 3, 1, 2, SQUARE_FLOOR);
		if (roomnum & 8)
		    rect(def, x * 8 + 3, y * 8 + 6, 2, 1, SQUARE_FLOOR);
		if (roomnum & 16)
		    rect(def, x * 8 + 3, y * 8 + 3, 2, 2, SQUARE_FLOOR);
	    }
	}
    }

    // Determine our links.
    // Each room has four links.
    // All links are bidirectional
    // This makes a 4-valence random graph.
    // We want it connected and not trivial.
    PTRLIST<int>		connectionlist[4];
    PTRLIST<int>		linkid[4];
    PTRLIST<int>		numconnect;
    int				numroom = maze_width * maze_height;
    int				linkid_base = 200;

    for (int i = 0; i < 4; i++)
    {
	connectionlist[i].resize(numroom);
	connectionlist[i].constant(-1);
	linkid[i].resize(numroom);
	linkid[i].constant(-1);
    }
    numconnect.resize(numroom);
    numconnect.constant(0);

    // Build the longest path.
    PTRLIST<int>	roompath;
    int			numrooms = maze_width * maze_height;
    for (int i = 1; i < numrooms-1; i++)
    {
	roompath.append(i);
    }
    roompath.shuffle();
    int			lastroom = 0;
    numconnect(0) = 1;
    for (int i = 0; i < roompath.entries(); i++)
    {
	connectionlist[0](lastroom) = roompath(i);
	connectionlist[1](roompath(i)) = lastroom;
	linkid[0](lastroom) = linkid_base;
	linkid[1](roompath(i)) = linkid_base;
	linkid_base++;
	numconnect(roompath(i)) = 2;

	lastroom = roompath(i);
    }
    connectionlist[0](lastroom) = numroom-1;
    connectionlist[0](numroom-1) = lastroom;
    linkid[0](lastroom) = linkid_base;
    linkid[0](numroom-1) = linkid_base;
    linkid_base++;
    numconnect(numroom-1) = 1;

    while (1)
    {
	PTRLIST<int>		potentialroom;
	bool			connect = false;

	for (int i = 0; i < numroom; i++)
	{
	    if (numconnect(i) < 4)
		potentialroom.append(i);
	}
	potentialroom.shuffle();
	// Connect all pairs!
	for (int i = 0; i < potentialroom.entries()-1; i += 2)
	{
	    int		a = potentialroom(i);
	    int		b = potentialroom(i+1);
	    if ( (a == 0 && b == numroom-1) ||
		 (b == 0 && a == numroom-1) )
	    {
		// Skip!
		continue;
	    }
	    connect = true;

	    connectionlist[numconnect(a)](a) = b;
	    connectionlist[numconnect(b)](b) = a;
	    linkid[numconnect(a)](a) = linkid_base;
	    linkid[numconnect(b)](b) = linkid_base;
	    linkid_base++;
	    numconnect(a)++;
	    numconnect(b)++;
	}

	if (!connect)
	    break;
    }

    // Now build the connections!
    for (int i = 0; i < 4; i++)
    {
	for (int j = 0; j < numconnect.entries(); j++)
	{
	    int		mx = j % maze_width;
	    int		my = j / maze_width;
	    int		dx = i & 1;
	    int		dy = i & 2 ? 1 : 0;

	    if (numconnect(j) <= i)
	    {
		set(def, mx * 8 + 2 + 3*dx, my * 8 + 2 + 3*dy, SQUARE_FLOOR);
		continue;
	    }
	    int		linkno = linkid[i](j);

	    assert(linkno > 0);
	    addLink(def, mx * 8 + 2 + 3*dx, my * 8 + 2 + 3*dy, 
		    linkno);
	}
    }


    // Entrance room 0, exit room end.
    set(def, 0, 3, SQUARE_FLOOR);
    addLink(def, 0, 3, 100);
    set(def, 1, 3, SQUARE_FLOOR);
    set(def, 2, 3, SQUARE_FLOOR);

    set(def, 8 * maze_width - 1 - 0, 8 * maze_height - 4, SQUARE_FLOOR);
    addLink(def, 8 * maze_width - 1, 8 * maze_height - 4, 100);
    set(def, 8 * maze_width - 1 - 1, 8 * maze_height - 4, SQUARE_FLOOR);
    set(def, 8 * maze_width - 1 - 2, 8 * maze_height - 4, SQUARE_FLOOR);

#ifdef STILL_TESTING
    // Validate our links.
    {
	MAP		*map;
	map = new MAP(*(ROOM_DEF *)def);
	printf("Validating teleMaze...\n");

	int failed = map->validateLinks(false);
	if (failed)
	{
	    printf("Failed building %d failures\n",
		    failed);
	    assert(!"Failed to validate!");
	}
	delete map;
    }
#endif

    return def;
}

class
LADDERBUILDER : public BUILD	// OOH!  Evil :>
{
public:
    LADDERBUILDER(int w, int h);

    int		 idxFromCoord(int level, int x, int y) const
    { assert(valid(x, y)); return level * maze_width * maze_height + y * maze_width + x; }
    void	 coordFromIdx(int idx, int &level, int &x, int &y) const
    {
	level = idx / (maze_width * maze_height);
	y = (idx / maze_width) % maze_height;
	x = idx % maze_width;
	assert(valid(x,y));
	assert(level >= 0 && level < 3);
    }

    bool	 valid(int x, int y) const
    { if (x < 0 || y < 0) return false;
      if (x >= maze_width || y >= maze_height) return false;
      return true; }

    void	 connectRooms(PTRLIST<int> &connect,
			    int level, int a, int b);
    void	 connectLevels(PTRLIST<int> &connect,
			    int aidx, int bidx);

    ROOM_DEF_NC	*def, *up, *low;
    ROOM_DEF_NC	*levels[3];
    int		 maze_width, maze_height;
};

static int
findClass(PTRLIST<int> &connect, int idx)
{
    int		c = connect(idx);

    if (c < 0)
	return c;

    if (c == idx)
	return c;

    c = findClass(connect, c);
    connect(idx) = c;

    return c;
}

static void
unionClass(PTRLIST<int> &connect, int aidx, int bidx)
{
    int ca = findClass(connect, aidx);
    int cb = findClass(connect, bidx);
    if (ca == cb)
	return;

    if (ca < cb)
	connect(cb) = ca;
    else
	connect(ca) = cb;
    findClass(connect, aidx);
    findClass(connect, bidx);
    assert(findClass(connect, aidx) == findClass(connect, bidx));
}

LADDERBUILDER::LADDERBUILDER(int w, int h)
{
    maze_width = w;
    maze_height = h;

    def = allocRoom(maze_width * 3 + 1, maze_height * 3 + 1);
    up = allocRoom(def->size.x, def->size.y);
    low = allocRoom(def->size.x, def->size.y);

    levels[0] = low;
    levels[1] = def;
    levels[2] = up;

    up->name = strdup("ladder_gen_up");
    up->down = strdup("ladder_gen");
    low->name = strdup("ladder_gen_down");
    low->up = strdup("ladder_gen");
    def->up = strdup(up->name);
    def->down = strdup(low->name);

    // Build our blank cells.
    fill(up, SQUARE_MOUNTAINS);
    fill(def, SQUARE_WALL);
    fill(low, SQUARE_SOLIDEARTH);
    for (int y = 0; y < maze_height; y++)
    {
	for (int x = 0; x < maze_width; x++)
	{
	    rect(up, x*3+1, y*3+1, 2, 2, SQUARE_GRASS);
	    rect(def, x*3+1, y*3+1, 2, 2, SQUARE_FLOOR);
	    set(low, x*3+2, y*3+2, SQUARE_ROCKGROUND);
	}
    }

    // Build our connectivity map
    PTRLIST<int>	connect;
    for (int i = 0; i < maze_width * maze_height * 3; i++)
	connect.append(i);

    // We want to create a bunch of small disjoint groups.
    for (int level = 0; level < 3; level++)
    {
	PTRLIST<int>	roomused;
	PTRLIST<int>	roomlist;

	roomused.resize(maze_width * maze_height);
	roomused.constant(0);

	for (int i = 0; i < roomused.entries(); i++)
	    roomlist.append(i);
	roomlist.shuffle();

	for (int i = 0; i < roomlist.entries(); i++)
	{
	    if (roomused(i))
		continue;

	    PTRLIST<int>	tetra;
	    tetra.append(i);
	    roomused(i) = true;
	    for (int j = 0; j < 3; j++)
	    {
		PTRLIST<int>	cuta, cutb;
		PTRLIST<int>	backupa, backupb;

		for (int k = 0; k < tetra.entries(); k++)
		{
		    int		dx, dy;
		    int		x, y, templvl;

		    int		a, b;

		    a = tetra(k);

		    coordFromIdx(a, templvl, x, y);
		    FORALL_4DIR(dx, dy)
		    {
			if (!valid(x+dx, y+dy))
			    continue;

			b = idxFromCoord(0, x+dx, y+dy);
			backupa.append(a);
			backupb.append(b);
			if (!roomused(b))
			{
			    cuta.append(a);
			    cutb.append(b);
			}
		    }
		}

		// Make the cut!
		if (cuta.entries())
		{
		    int		c = rand_choice(cuta.entries());
		    connectRooms(connect, level, cuta(c), cutb(c));
		    roomused(cutb(c)) = true;

		    tetra.append(cutb(c));
		}
		else
		{
		    // We've been isolated.  No problem, make a single
		    // cut and return
		    int		c = rand_choice(backupa.entries());
		    connectRooms(connect, level, backupa(c), backupb(c));
		    roomused(backupb(c)) = true;

		    break;
		}
	    }
	}
    }

    replace(up, SQUARE_GRASS, SQUARE_FOREST, 15);

    // Cut every pair of levels that differs, sorting randomly
    {
	PTRLIST<int>		cutlist;

	for (int i = 0; i < 2 * maze_width * maze_height; i++)
	    cutlist.append(i);

	cutlist.shuffle();

	for (int i = 0; i < cutlist.entries(); i++)
	{
	    int		a, b;
	    a = cutlist(i);
	    b = a + maze_width * maze_height;
	    if (findClass(connect, a) != findClass(connect, b)
		)
//		|| rand_chance(5))
	    {
		// Need to connect!
		connectLevels(connect, a, b);
	    }
	}
    }

    // We may still not be connected, but likely are
    // To be safe, we connect all within each level.
    {
	PTRLIST<int>		cutlist;
	for (int i = 0; i < 3 * maze_width * maze_height; i++)
	    cutlist.append(i);
	cutlist.shuffle();

	for (int i = 0; i < cutlist.entries(); i++)
	{
	    int		x, y, level;

	    int a = cutlist(i);
	    coordFromIdx(a, level, x, y);
	    int		dx, dy;

	    FORALL_4DIR(dx, dy)
	    {
		if (!valid(x+dx, y+dy))
		    continue;

		int b = idxFromCoord(level, x+dx, y+dy);
		if (findClass(connect, a) != findClass(connect, b))
		{
		    connectRooms(connect, level,
					idxFromCoord(0, x, y),
					idxFromCoord(0, x+dx, y+dy));
		}
	    }
	}
    }

    // Validate all rooms are connected!
#ifdef STILL_TESTING
    {
	int		baseclass = findClass(connect, 0);
	int		failed = 0;
	for (int i = 0; i < maze_width * maze_height * 3; i++)
	{
	    if (findClass(connect, i) != baseclass)
	    {
		int		l, x, y;
		coordFromIdx(i, l, x, y);
		printf("Unconnected room %d,%d level %d, class %d\n",
			x, y, l, findClass(connect, i));
		failed++;
	    }
	}

	assert(!failed);
    }
#endif

    // Build our entrance and exit
    int		ey;

    ey = randomSquareInColumn(def, SQUARE_FLOOR, 1);
    set(def, 0, ey, SQUARE_DOOR);
    addLink(def, 0, ey, 100);

    ey = randomSquareInColumn(def, SQUARE_FLOOR, def->size.x-2);
    set(def, def->size.x-1, ey, SQUARE_FLOOR);
    addLink(def, def->size.x-1, ey, 100);
    
    registerDef(low);
    registerDef(up);
}

void
LADDERBUILDER::connectLevels(PTRLIST<int> &connect,
			int aidx, int bidx)
{
    int		ax, ay, level;

    coordFromIdx(aidx, level, ax, ay);

    switch (level)
    {
	case 0:
	{
	    addItem(levels[0], ax*3+2, ay*3+2, ITEM_LADDERUP);
	    addItem(levels[1], ax*3+2, ay*3+2, ITEM_LADDERDOWN);
	    break;
	}
	case 1:
	{
	    addItem(levels[1], ax*3+1, ay*3+1, ITEM_LADDERUP);
	    addItem(levels[2], ax*3+1, ay*3+1, ITEM_LADDERDOWN);
	    break;
	}

	case 2:
	{
	    assert(!"Invalid level to connect!");
	    break;
	}
    }

    // Connect the sets.
    unionClass(connect, aidx, bidx);
}

void
LADDERBUILDER::connectRooms(PTRLIST<int> &connect,
			int level, int a, int b)
{
    int		aidx, bidx;

    aidx = a + maze_width * maze_height * level;
    bidx = b + maze_width * maze_height * level;

    int		ax, ay, bx, by;
    int		dx, dy, ilevel;

    coordFromIdx(aidx, ilevel, ax, ay);
    assert(ilevel == level);
    coordFromIdx(bidx, ilevel, bx, by);
    assert(ilevel == level);

    dx = SIGN(bx-ax);
    dy = SIGN(by-ay);

    switch (level)
    {
	case 0:
	{
	    // Straight connection.
	    for (int x = ax * 3+2, y = ay * 3+2;
		 x != (bx*3+2) || y != (by*3+2);
		 x += dx, y += dy)
	    {
		set(levels[level], x, y, SQUARE_ROCKGROUND);
	    }
	    break;
	}
	case 1:
	{
	    // Either a door or a hallway.
	    if (rand_chance(25))
	    {
		int	doorbias = rand_choice(2) + 1;
		int	doorx = (ax + (dx==1)) * 3 + (!dx)*doorbias;
		int 	doory = (ay + (dy==1)) * 3 + (!dy)*doorbias;
		assert(doorx > 0);
		assert(doory > 0);
		set(levels[level], 
			doorx,
			doory,
			SQUARE_DOOR);
	    }
	    else
	    {
		// Hallway.
		for (int x = ax * 3+1, y = ay * 3+1;
		     x != (bx*3+1) || y != (by*3+1);
		     x += dx, y += dy)
		{
		    set(levels[level], x, y, SQUARE_FLOOR);
		    set(levels[level], x+!dx, y+!dy, SQUARE_FLOOR);
		}
	    }
	    break;
	}

	case 2:
	{
	    // Always hallway.
	    for (int x = ax * 3+1, y = ay * 3+1;
		 x != (bx*3+1) || y != (by*3+1);
		 x += dx, y += dy)
	    {
		set(levels[level], x, y, SQUARE_GRASS);
		set(levels[level], x+!dx, y+!dy, SQUARE_GRASS);
	    }
	    break;
	}
    }

    // Connect the sets.
    unionClass(connect, aidx, bidx);
}


ROOM_DEF_NC *
BUILD::buildLadderMaze(int maze_width, int maze_height)
{
    LADDERBUILDER	builder(maze_width, maze_height);

    return builder.def;
}

ROOM_DEF_NC *
BUILD::buildTrollPrison()
{
    ROOM_DEF_NC		*def;

    def = allocRoom(19, 9);
    fill(def, SQUARE_WALL);

    // Entrance
    set(def, 0, 4, SQUARE_FLOOR);
    addLink(def, 0, 4, 100);

    // Exit
    set(def, 18, 5, SQUARE_FLOOR);
    addLink(def, 18, 5, 100);

    rect(def, 1, 1, 4, 7, SQUARE_FLOOR);
    for (int i = 0; i < 2; i++)
    {
	addItem(def, rand_range(1, 4), rand_range(1, 7), ITEM_BONES);
    }

    set(def, 4, 0, SQUARE_WALL_LEVER_DOWN);
    addLink(def, 4, 0, 1);
    set(def, 5, 1, SQUARE_PORTICULLIS_DOWN);
    addLink(def, 5, 1, 1);

    set(def, 5, 5, SQUARE_WALL_LEVER_DOWN);
    addLink(def, 5, 5, 2);
    set(def, 5, 4, SQUARE_PORTICULLIS_DOWN);
    addLink(def, 5, 4, 2);

    set(def, 4, 8, SQUARE_WALL_LEVER_DOWN);
    addLink(def, 4, 8, 3);
    set(def, 5, 7, SQUARE_PORTICULLIS_DOWN);
    addLink(def, 5, 7, 3);

    // Now see which connects up!
    int		rightexit_y, rightexit;
    const int	possibleexit[3] = { 1, 4, 7 };
    rightexit = rand_choice(3);
    rightexit_y = possibleexit[rightexit];

    for (int i = 0; i < 3; i++)
    {
	rect(def, 6, possibleexit[i], 8, 1, SQUARE_FLOOR);
	addMob(def, 8, possibleexit[i], MOB_TROLL);
	addMob(def, 10, possibleexit[i], MOB_TROLL);
	set(def, 14, possibleexit[i], SQUARE_PORTICULLIS_DOWN);
    }

    // Tie our correct exit to the port!
    addLink(def, 14, rightexit_y, rightexit+1);

    // Build the collector hallway
    rect(def, 15, 1, 1, 7, SQUARE_FLOOR);

    // Build the exit
    rect(def, 16, 5, 2, 1, SQUARE_FLOOR);

    // Add an extra lever to shut the doors!
    set(def, 17, 6, SQUARE_WALL_LEVER_DOWN);
    addLink(def, 17, 6, rightexit+1);

    return def;
}

bool 
BUILD::drawRiver(ROOM_DEF_NC *def,
			const ROOMDATA<float> &height,
			int x, int y,
			SQUARE_NAMES water,
			SQUARE_NAMES dest)
{
    int		dx, dy;
    int		cdx, cdy;
    float	dist = height(x, y);

    cdx = 1;
    cdy = 0;
    while (get(def, x, y) != dest)
    {
	bool		found = false;
#if 1
	float	ndist;
	int	choice = 0;
	FORALL_4DIR(dx, dy)
	{
	    if (height(x+dx, y+dy) > dist)
	    {
		choice++;
		if (!rand_choice(choice))
		{
		    cdx = dx;
		    cdy = dy;
		    ndist = height(x+dx, y+dy);
		}
	    }
	}
	if (choice)
	{
	    found = true;
	    dist = ndist;
	}
#else
	// Steepest descent.
	FORALL_4DIR(dx, dy)
	{
	    if (height(x+dx, y+dy) > dist)
	    {
		found = true;
		dist = height(x+dx, y+dy);
		cdx = dx;
		cdy = dy;
	    }
	}
#endif
	x += cdx;
	y += cdy;
	if (get(def, x, y) == dest)
	    break;

	if (!found)
	{
	    // Gah!
	    // Make a pool, ensure we are not in mountains
	    rect(def, x-1, y-1, 3, 3, water);

	    // FInd closest destination and draw a line right there.
	    int		tx = x, ty = y;
	    findClosest(def, dest, tx, ty);

	    assert(get(def, tx, ty) == dest);

	    while (x != tx || y != ty)
	    {
		int		dx = SIGN(tx-x);
		int		dy = SIGN(ty-y);
		x += dx;
		set(def, x, y, water);
		y += dy;
		set(def, x, y, water);
	    }

	    return true;
	}

	set(def, x, y, water);
    }
    return true;
}

void
BUILD::findClosest(ROOM_DEF_NC *def, SQUARE_NAMES goal,
		    int &sx, int &sy)
{
    int		r = 0;

    for (r = 0; r < MAX(def->size.x, def->size.y); r++)
    {
	for (int sr = 0; sr <= r; sr++)
	{
	    int		 x = sx;
	    int		 y = sy;
	    if (get(def, x-r, y-sr) == goal)
	    {
		sx = x-r;
		sy = y-sr;
		return;
	    }
	    if (get(def, x-r, y+sr) == goal)
	    {
		sx = x-r;
		sy = y+sr;
		return;
	    }
	    if (get(def, x+r, y-sr) == goal)
	    {
		sx = x+r;
		sy = y-sr;
		return;
	    }
	    if (get(def, x+r, y+sr) == goal)
	    {
		sx = x+r;
		sy = y+sr;
		return;
	    }
	    if (get(def, x-sr, y-r) == goal)
	    {
		sx = x-sr;
		sy = y-r;
		return;
	    }
	    if (get(def, x-sr, y+r) == goal)
	    {
		sx = x-sr;
		sy = y+r;
		return;
	    }
	    if (get(def, x+sr, y-r) == goal)
	    {
		sx = x+sr;
		sy = y-r;
		return;
	    }
	    if (get(def, x+sr, y+r) == goal)
	    {
		sx = x+sr;
		sy = y+r;
		return;
	    }
	}
    }
}

void
BUILD::findFarthest(ROOM_DEF_NC *def, SQUARE_NAMES goal,
		    int &rsx, int &rsy)
{
    int		r = 0;
    int		sx = rsx, sy = rsy;

    for (r = 0; r < MAX(def->size.x, def->size.y); r++)
    {
	for (int sr = 0; sr <= r; sr++)
	{
	    int		 x = sx;
	    int		 y = sy;
	    if (get(def, x-r, y-sr) == goal)
	    {
		rsx = x-r;
		rsy = y-sr;
	    }
	    if (get(def, x-r, y+sr) == goal)
	    {
		rsx = x-r;
		rsy = y+sr;
	    }
	    if (get(def, x+r, y-sr) == goal)
	    {
		rsx = x+r;
		rsy = y-sr;
	    }
	    if (get(def, x+r, y+sr) == goal)
	    {
		rsx = x+r;
		rsy = y+sr;
	    }
	    if (get(def, x-sr, y-r) == goal)
	    {
		rsx = x-sr;
		rsy = y-r;
	    }
	    if (get(def, x-sr, y+r) == goal)
	    {
		rsx = x-sr;
		rsy = y+r;
	    }
	    if (get(def, x+sr, y-r) == goal)
	    {
		rsx = x+sr;
		rsy = y-r;
	    }
	    if (get(def, x+sr, y+r) == goal)
	    {
		rsx = x+sr;
		rsy = y+r;
	    }
	}
    }
}

void
BUILD::buildTruePermutation(PTRLIST<int> &list, int len)
{
    list.resize(0);
    for (int i = 0; i < len; i++)
	list.append(i);
    list.shuffle();
    // Small lists can't be fixed.
    if (list.entries() < 2)
	return;
    // We may have self-references.
    // RUn through everything.  If we find an illegal, swap with
    // someone who won't be unhappy coming into our spot
    for (int i = 0; i < list.entries(); i++)
    {
	if (list(i) == i)
	{
	    // Uh-oh!  Don't want to get ourself as a prize :>
	    // By construction, we can always swap ourselves with
	    // a different random entry since it must not equal
	    // us and we won't equal its rank!
	    int		j = rand_choice(list.entries()-1);
	    if (j >= i)
		j++;
	    list.swapEntries(i, j);
	}
    }
}
