#include "map.h"

#include "gfxengine.h"
#include "atlas.h"
#include "text.h"
#include <stdio.h>


MAP::MAP(const ROOM_DEF &def)
{
    myDef = &def;
    myWidth = def.size.x;
    myHeight = def.size.y;

    mySquares = new SQUARE_NAMES[myWidth * myHeight];
    myItems = new ITEM_NAMES[myWidth * myHeight];
    myMobCache = new MOB *[myWidth * myHeight];
    myFlags = new u8[myWidth * myHeight];

    myName.reference(def.name);

    myAtlas = 0;
    myMapId = -1;

    memset(myMobCache, 0, myWidth * myHeight * sizeof(MOB *));

    resetAll();
}

MAP::~MAP()
{
    delete [] mySquares;
    delete [] myItems;
    delete [] myMobCache;
    delete [] myFlags;
}

void
MAP::resetAll()
{
    int		x, y;

    memset(mySquares, 0, myWidth * myHeight * sizeof(SQUARE_NAMES));
    memset(myItems, 0, myWidth * myHeight * sizeof(ITEM_NAMES));
    memset(myFlags, 0, myWidth * myHeight * sizeof(u8));

    for (int mi = myMobList.entries(); mi --> 0; )
    {
	MOB	*m = myMobList(mi);
	if (!m->isAvatar())
	{
	    delete m;
	    // Do not do removeAt as that is done by the mob deletion!
	}
    }

    FORALL_XY(x, y)
    {
	mySquares[x + y * myWidth] = defn().squarelist[x + y * myWidth];
    }

    for (const PT2 *flag = defn().squareflaglist; flag && flag->data; flag++)
    {
	setFlag((FLAG_NAMES) flag->data, flag->x, flag->y, true);
    }

    // Set our flags from global flags
    FORALL_XY(x, y)
    {
	setFlag((FLAG_NAMES) defn().globalmapflags, x, y, true);
    }

    for (const PT2 *item = defn().itemlist; item && item->data; item++)
    {
	myItems[item->x + item->y * myWidth] = (ITEM_NAMES) item->data;
    }

    for (const PT2 *mob = defn().moblist; mob && mob->data; mob++)
    {
	MOB		*m = new MOB((MOB_NAMES) mob->data);

	m->moveTo(buildPos(mob->x, mob->y));
    }
}

void
MAP::heartbeat(int phase)
{
    // Downwards as movement may delete mobs!
    for (int m = myMobList.entries(); m --> 0; )
    {
	if (myMobList(m)->heartbeat(phase))
	{
	    myMobList(m)->doAI();
	}
    }

    int		x, y;

    // Spread forest fires.
    // We do this first so we don't spread them on their
    // first run.
    FORALL_XY(x, y)
    {
	POS	p = buildPos(x, y);
	DIR	dir;
	POSLIST	setafire;		// Delay lighting to avoid
					// instant propagation as far
					// as possible.
	if (p.defn().burning)
	{
	    FORALL_DIR(dir)
	    {
		POS	g = p.delta(dir);

		if (g.defn().burn != SQUARE_NONE)
		{
		    // Potential to spread!
		    if (rand_chance(20))
			setafire.append(g);
		}
	    }

	    if (rand_chance(20))
	    {
		p.setTile((SQUARE_NAMES) p.defn().burned);
	    }
	}

	for (int i = 0; i < setafire.entries(); i++)
	{
	    setafire(i).setTile((SQUARE_NAMES) setafire(i).defn().burn);
	}
    }

    // Apply fire effects!
    FORALL_XY(x, y)
    {
	if (getFlag(FLAG_FIRE, x, y))
	{
	    if (getFlag(FLAG_FIREFRESH, x, y))
	    {
		setFlag(FLAG_FIREFRESH, x, y, false);
	    }
	    else
	    {
		// End the fire!
		setFlag(FLAG_FIRE, x, y, false);
		setFlag(FLAG_BURNED, x, y, true);

		if (glb_squaredefs[getTile(x, y)].burn != SQUARE_NONE)
		{
		    setTile(x, y, (SQUARE_NAMES)glb_squaredefs[getTile(x, y)].burn);
		}
		POS	p = buildPos(x, y);
		if (glb_itemdefs[p.item()].burnable)
		    p.setItem(ITEM_NONE);		// Oops!
	    }
	}

    }
}

void
MAP::clearMessages()
{
    for (int m = 0; m < myMobList.entries(); m++)
    {
	myMobList(m)->clearMessages();
    }
}

MAP::iterator::iterator()
{
    myMap = 0;
    x = y = 0;
}

MAP::iterator::iterator(const MAP *map)
{
    myMap = map;
    x = y = 0;
}

void
MAP::iterator::advance()
{
    if (y >= myMap->height())
	return;

    x++;
    if (x >= myMap->width())
    {
	x = 0;
	y++;
    }
}

POS
MAP::iterator::pos() const
{
    return myMap->buildPos(x, y);
}

bool
MAP::iterator::done() const
{
    if (!myMap)
	return true;
    return y >= myMap->height();
}

MAP::iterator
MAP::begin() const
{
    MAP::iterator	it(this);

    return it;
}

MOB *
MAP::findMobByType(MOB_NAMES mob) const
{
    for (int m = 0; m < myMobList.entries(); m++)
    {
	if (!myMobList(m))
	    continue;

	if (myMobList(m)->getDefinition() == mob)
	    return myMobList(m);
    }
    return 0;
}

int
MAP::findMapLink(int xpos, int ypos) const
{
    if (xpos == 1)
	xpos = width()-1;
    if (ypos == 1)
	ypos = height()-1;

    for (int i = 0; defn().linklist[i].data >= 0; i++)
    {
	if ((ypos == -1 && defn().linklist[i].x == xpos) ||
	    (xpos == -1 && defn().linklist[i].y == ypos))
	{
	    // Check the link number
	    if (defn().linklist[i].data == 100)
	    {
		if (xpos == -1)
		    return defn().linklist[i].x;
		if (ypos == -1)
		    return defn().linklist[i].y;
	    }
	}
    }
    cerr << "No map link found" << endl;
    return 0;
}


void
MAP::makePosValid(POS &p)
{
    if (validPos(p.x(), p.y()))
	return;

    MAP		*adj;
    if (p.x() < 0)
    {
	if (defn().west)
	{
	    adj = atlas()->findMap(defn().west);
	    if (adj)
	    {
		int		ay = adj->findMapLink(1, -1);
		int		my = findMapLink(0, -1);
		p = adj->buildPos(p.x()+adj->width(), p.y() + ay - my);
		adj->makePosValid(p);
		return;
	    }
	}
    }
    if (p.x() >= width())
    {
	if (defn().east)
	{
	    adj = atlas()->findMap(defn().east);
	    if (adj)
	    {
		int		ay = adj->findMapLink(0, -1);
		int		my = findMapLink(1, -1);
		p = adj->buildPos(p.x()-width(), p.y() + ay - my);
		adj->makePosValid(p);
		return;
	    }
	}
    }
    if (p.y() < 0)
    {
	if (defn().north)
	{
	    adj = atlas()->findMap(defn().north);
	    if (adj)
	    {
		int		ax = adj->findMapLink(-1, 1);
		int		mx = findMapLink(-1, 0);
		p = adj->buildPos(p.x() + ax - mx, p.y()+adj->height());
		adj->makePosValid(p);
		return;
	    }
	}
    }
    if (p.y() >= height())
    {
	if (defn().south)
	{
	    adj = atlas()->findMap(defn().south);
	    if (adj)
	    {
		int		ax = adj->findMapLink(-1, 0);
		int		mx = findMapLink(-1, 1);
		p = adj->buildPos(p.x() + ax - mx, p.y()-height());
		adj->makePosValid(p);
		return;
	    }
	}
    }

    if (validPos(p.x(), p.y()))
	return;

    // Should check for portals here.
    p.myMap = 0;
}

void
MAP::setAtlas(ATLAS *atlas, int mapid)
{
    myAtlas = atlas;
    myMapId = mapid;
}

void
MAP::clear(SQUARE_NAMES square)
{
    for (int y = 0; y < height(); y++)
    {
	for (int x = 0; x < width(); x++)
	{
	    setTile(x, y, square);
	}
    }
}

bool
MAP::validPos(int x, int y) const
{
    if (x < 0 || x >= width())
	return false;
    if (y < 0 || y >= height())
	return false;

    return true;
}

SQUARE_NAMES
MAP::getTile(int x, int y) const
{
    if (!validPos(x, y))
	return SQUARE_NONE;

    return mySquares[x + y * width()];
}

ITEM_NAMES
MAP::getItem(int x, int y) const
{
    if (!validPos(x, y))
	return ITEM_NONE;

    return myItems[x + y * width()];
}

void
MAP::setTile(int x, int y, SQUARE_NAMES square) 
{
    assert(validPos(x, y));

    mySquares[x + y * width()] = square;
}


void
MAP::setItem(int x, int y, ITEM_NAMES item) 
{
    assert(validPos(x, y));

    myItems[x + y * width()] = item;
}

void
MAP::getLinks(int x, int y, PTRLIST<int> &links) const
{
    for (int i = 0; defn().linklist[i].data >= 0; i++)
    {
	if (defn().linklist[i].x == x &&
	    defn().linklist[i].y == y)
	{
	    links.append(defn().linklist[i].data);
	}
    }
}

void
MAP::findLinks(int link, POSLIST &plist) const
{
    for (int i = 0; defn().linklist[i].data >= 0; i++)
    {
	if (defn().linklist[i].data == link)
	{
	    plist.append(buildPos(defn().linklist[i].x,
				  defn().linklist[i].y));
	}
    }
}

bool
MAP::getFlag(FLAG_NAMES flag, int x, int y) const
{
    if (!validPos(x, y))
	return SQUARE_NONE;

    return (myFlags[x + y * width()] & flag) ? true : false;
}

void
MAP::resetFlag(FLAG_NAMES flag, bool val) 
{
    int		x, y;

    if (val)
    {
	FORALL_XY(x, y)
	{
	    myFlags[x + y * width()] |= flag;
	}
    }
    else
    {
	FORALL_XY(x, y)
	{
	    myFlags[x + y * width()] &= ~((int)flag);
	}
    }
}

void
MAP::setFlag(FLAG_NAMES flag, int x, int y, bool val) 
{
    assert(validPos(x, y));

    if (val)
	myFlags[x + y * width()] |= flag;
    else
	myFlags[x + y * width()] &= ~((int)flag);
}

MOB *
MAP::getMob(int x, int y ) const
{
    if (!validPos(x, y))
	return 0;

    return myMobCache[x + y * width()];
}

void
MAP::setMob(int x, int y, MOB *mob)
{
    assert (validPos(x, y));

    myMobCache[x + y * width()] = mob;
}

void
MAP::registerMob(MOB *mob)
{
    myMobList.append(mob);
    assert(mob->pos().map() == this);

    setMob(mob->pos().myX, mob->pos().myY, mob);
}

void
MAP::deregisterMob(MOB *mob)
{
    myMobList.removePtr(mob);

    assert(getMob(mob->pos().myX, mob->pos().myY) == mob);
    if (getMob(mob->pos().myX, mob->pos().myY) == mob)
    {
	setMob(mob->pos().myX, mob->pos().myY, 0);
    }
}

void
MAP::updateMobCache(MOB *mob, POS oldpos, POS newpos)
{
    if (getMob(oldpos.myX, oldpos.myY) == mob)
	setMob(oldpos.myX, oldpos.myY, 0);

    setMob(newpos.myX, newpos.myY, mob);
}

POS
MAP::buildPos(int x, int y) const
{
    POS		pos;

    pos.myX = x;
    pos.myY = y;
    pos.myMap = (MAP *)this;

    return pos;
}

static void
renderTile(int sx, int sy, POS pos, bool reveal)
{
    if (sx < 0) return;
    if (sx >= SCR_WIDTH/4) return;
    if (sy < 0) return;
    if (sy >= SCR_HEIGHT/4) return;

    if (!pos.getFlag(FLAG_FOV))
    {
	gfx_settile(0, sx, sy, TILE_BLEEDABLE_OPAQUE);
	gfx_settile(1, sx, sy, TILE_NONE);
	gfx_settile(2, sx, sy, TILE_NONE);
	gfx_settile(3, sx, sy, TILE_BLEEDABLE_TRANS);
	return;
    }

    if (pos.getFlag(FLAG_SHADOW))
	gfx_settile(3, sx, sy, TILE_SHADOW);
    else if (pos.getFlag(FLAG_FIRE))
	gfx_settile(3, sx, sy, TILE_FIRE);
    else if (pos.getFlag(FLAG_DARK))
	gfx_settile(3, sx, sy, TILE_SHADOW);
    else
	gfx_settile(3, sx, sy, TILE_BLEEDABLE_TRANS);

    MOB		*mob;

    if (reveal && pos.defn().reveal != TILE_NONE)
	gfx_settile(0, sx, sy, (TILE_NAMES) pos.defn().reveal);
    else
	gfx_settile(0, sx, sy, (TILE_NAMES) pos.defn().tile);

    gfx_settile(1, sx, sy, (TILE_NAMES) glb_itemdefs[pos.item()].tile);

    mob = pos.mob();
    if (mob)
    {
	if (!reveal && mob->intrinsic().get(INTRINSIC_INVIS))
	{
	    // You can see your self.
	    if (mob->isAvatar())
		gfx_settile(2, sx, sy, TILE_INVIS);
	}
	else
	    gfx_settile(2, sx, sy, (TILE_NAMES) mob->defn().tile);

	const char *msg = mob->getMessage();
	const char *emote = mob->getEmote();
	bool		hasemote = (emote && *emote);
	bool		hasmsg = (msg && *msg);
	if (hasemote || hasmsg)
	{
	    int		tx = sx * 4 + 3;
	    int		ty = sy * 4;
	    gfx_printstring(tx, ty-1, "/", hasemote ? ATTR_EMOTE : ATTR_SPEECH);
	    if (hasemote)
		gfx_printstring(tx+1-strlen(emote)/2, ty-2, emote, ATTR_EMOTE);
	    if (hasmsg)
		gfx_justifystring(tx+1, ty-2-hasemote, JUST_CENTER, JUST_BOTTOM, msg, ATTR_SPEECH);
	}
    }
    else
    {
	// We intentionaly put this on the mob layer as we
	// want it hidden by walking over it.
	if (pos.getFlag(FLAG_BURNED) && !pos.defn().burning)
	    gfx_settile(2, sx, sy, TILE_BURNED);
	else
	    gfx_settile(2, sx, sy, TILE_NONE);
    }
}

void
moveScreen(int &x, int &y, DIR dir, DIR face)
{
    switch ((dir.angle() + face.angle()) & 3)
    {
	case 0:
	    y--;
	    break;
	case 1:
	    x--;
	    break;
	case 2:
	    y++;
	    break;
	case 3:
	    x++;
	    break;
    }
}

void
MAP::render(POS center, DIR facedir, MOB *avatar)
{
    // Draw our intrisics
    INTRINSIC_NAMES	intrinsic;
    int			statusx = 15;
    int			statusy = 2;
    FOREACH_INTRINSIC(intrinsic)
    {
	if (avatar->intrinsic().get(intrinsic))
	{
	    if (glb_intrinsicdefs[intrinsic].tile != TILE_NONE)
	    {
		gfx_settile(1, statusx, statusy, (TILE_NAMES)glb_intrinsicdefs[intrinsic].over);
		gfx_settile(0, statusx, statusy, (TILE_NAMES)glb_intrinsicdefs[intrinsic].tile);
		statusx++;
		if (statusx >= 20)
		{
		    statusy++;
		    statusx = 15;
		}
	    }
	}
    }

    while (statusy < 5)
    {
	gfx_settile(1, statusx, statusy, TILE_NONE);
	gfx_settile(0, statusx, statusy, TILE_NONE);
	statusx++;
	if (statusx >= 20)
	{
	    statusy++;
	    statusx = 15;
	}
    }
#if 0
    // This doesn't look good :<
    gfx_printstring(60, 8, "Name:", ATTR_NORMAL);
    gfx_printstring(60, 9, text_lookupvariable("name"), ATTR_NORMAL);
#endif

    bool		reveal = false;
    if (avatar && avatar->intrinsic().get(INTRINSIC_REVEAL))
	reveal = true;
#if 1
    DIR		dir;
    renderTile(7, 7, center, reveal);
    dir = DIR(0);
    FORALL_DIR(dir)
    {
	WAVEFRONT	front(center, dir);
	DIR		perp, backperp;
	int sx, sy;

	perp = dir.clock();
	backperp = dir.counterclock();
	sx = 7;
	sy = 7;

	for (int r = 1; r < SCR_RAD; r++)
	{
	    moveScreen(sx, sy, dir, facedir);
	    moveScreen(sx, sy, backperp, facedir);
	    int		wx = sx, wy = sy;
	    for (int i = 0; i < front.entries(); i++)
	    {
		renderTile(wx, wy, front(i), reveal);
		moveScreen(wx, wy, perp, facedir);
	    }
	    front.advance();
	}
    }
#else
    int		x, y;
    int		cx, cy;

    cx = center.x();
    cy = center.y();

    for (y = -1; y <= center.map()->height(); y++)
    for (x = -1; x <= center.map()->width(); x++)
    {
	int		sx, sy;
	
	sx = x - cx + 7;
	sy = y - cy + 7;

	POS		pos = center.map()->buildPos(x, y);

	renderTile(sx, sy, pos, reveal);
    }
#endif
}

u8 *
MAP::buildMiniMap() const
{
    u8	*rgb = new u8 [width() * height() * 3];
    int		x, y;

    FORALL_XY(x, y)
    {
	ATTR_NAMES		attr;

	attr = (ATTR_NAMES)glb_itemdefs[getItem(x, y)].minimap;
	if (attr == ATTR_NONE)
	    attr = (ATTR_NAMES)glb_squaredefs[mySquares[x + y * width()]].minimap;

	rgb[ (x + y * width()) * 3 + 0 ] = glb_attrdefs[attr].fg_b;
	rgb[ (x + y * width()) * 3 + 1 ] = glb_attrdefs[attr].fg_g;
	rgb[ (x + y * width()) * 3 + 2 ] = glb_attrdefs[attr].fg_r;
    }

    return rgb;
}

struct map_filterSquare
{
    map_filterSquare(SQUARE_NAMES square) { mySquare = square; }
    bool operator()(POS p) { return p.tile() == mySquare; }
    SQUARE_NAMES		mySquare;
};

struct map_filterItem
{
    map_filterItem(ITEM_NAMES item) { myItem = item; }
    bool operator()(POS p) { return p.item() == myItem; }
    ITEM_NAMES		myItem;
};

struct map_filterPassable
{
    map_filterPassable() { }
    bool operator()(POS p) { return p.defn().passable; }
};

POS
MAP::findRandomSquare(SQUARE_NAMES square) const
{
    map_filterSquare filter(square);

    return findRandomPos(filter);
}

POS
MAP::findRandomItem(ITEM_NAMES item) const
{
    map_filterItem filter(item);

    return findRandomPos(filter);
}

POS
MAP::findRandomPassable() const
{
    map_filterPassable filter;

    return findRandomPos(filter);
}

int
MAP::validateLinks(bool quiet) const
{
    int		x, y;
    PTRLIST<int>		links;
    PTRLIST<int>		linearpos;
    POSLIST			dests;
    int				failed = 0;
    FORALL_XY(x, y)
    {
	POS	p = buildPos(x, y);

	switch (p.tile())
	{
	    case SQUARE_WALL_LEVER_UP:
	    case SQUARE_WALL_LEVER_DOWN:
		p.findLinkedSquares(dests);
		links.clear();
		getLinks(x, y, links);
		if (links.entries() != 1)
		{
		    if (!quiet)
			printf("Lever at %d,%d has %d out links!\n",
				x, y, links.entries());
		    failed++;
		}

		// Verify no duplicate destinations.
		linearpos.clear();
		for (int i = 0; i < dests.entries(); i++)
		    linearpos.append( dests(i).x() + (dests(i).y() <<16) );
		linearpos.stableSort();
		for (int i = 1; i < dests.entries(); i++)
		{
		    if (linearpos(i) == linearpos(i-1))
		    {
			int	lx, ly;
			lx = linearpos(i) & 65535;
			ly = linearpos(i) >> 16;
			if (!quiet)
			    printf("Lever at %d,%d controls duplicate at %d,%d\n", x, y, lx, ly);
			failed++;
		    }
		}

		// Must be at least one other thing controlled, which
		// means dests >= 2
		if (dests.entries() < 2)
		{
		    if (!quiet)
			printf("Lever at %d,%d controls nothing!\n", x, y);
		    failed++;
		}

		break;

	    case SQUARE_FLOOR_TELE:
	    case SQUARE_GROUND_TELE:
		p.findLinkedSquares(dests);
		if (dests.entries() != 2)
		{
		    if (!quiet)
			printf("Teleporter at %d,%d has %d destinations!\n",
				x, y, dests.entries());
		    failed++;
		}
		else
		{
		    // Make sure we have something other than ourselves!
		    if ( dests(0).x() == x && dests(0).y() == y &&
			 dests(1).x() == x && dests(1).y() == y )
		    {
			if (!quiet)
			    printf("Teleporter at %d,%d goes only to itself!\n",
				    x, y);
			failed++;
		    }
		}
		break;
	}
    }

    return failed;
}
