#include "mob.h"
#include "map.h"
#include "dist.h"
#include "atlas.h"
#include "text.h"
#include "gfxengine.h"
#include "runes.h"

extern void readBook(const char *text);
extern void readBook(BUF text);
extern void fadeToBlack();
extern void fadeFromBlack();
extern void redrawWorld();

MOB::MOB(MOB_NAMES defn)
{
    myDefinition = defn;
    myPos = POS::invalidPos();
    myDiskSourceMap = 0;
    myLastPhase = 0;
    myFacing = DIR::north();
    myLastWalkDir = DIR::north();

    for (const char *i = MOB::defn().intrinsic; *i; i++)
    {
	intrinsic().set((INTRINSIC_NAMES) *i);
    }
}

MOB::~MOB()
{
    if (myPos.map())
	myPos.map()->deregisterMob(this);
}

void
MOB::moveTo(POS newpos)
{
    if (myPos.map())
    {
	if (newpos.map() != myPos.map())
	{
	    myPos.map()->deregisterMob(this);
	    if (newpos.map())
	    {
		myPos = newpos;
		newpos.map()->registerMob(this);
	    }
	}
	else
	{
	    myPos.map()->updateMobCache(this, myPos, newpos);
	}
    }
    else if (newpos.map())
    {
	myPos = newpos;
	newpos.map()->registerMob(this);
    }
    myPos = newpos;
}

bool
MOB::actionBump(DIR dir)
{
    POS  p = pos().delta(dir);

    if (p.mob())
    {
	return actionAttack(dir);
    }

    if (!canMoveIn(dir))
    {
	// Check for block actions...
	if (isAvatar())
	{
	    switch (p.tile())
	    {
		case SQUARE_DOOR:
		    actionEmote("*opens*");
		    p.setTile(SQUARE_DOOR_OPEN);
		    return true;
		case SQUARE_WALL_LEVER_DOWN:
		case SQUARE_WALL_LEVER_UP:
		    actionEmote("*flips*");
		    p.toggleLinkedSquares();
		    return true;
		default:
		    break;
	    }
	}
	// No slide, blocked
	return false;
    }

    return actionWalk(dir);
}

void
MOB::die(const char *death)
{
    MAP *map = glbAtlas()->findMap("home");
    POS p = map->findRandomItem(ITEM_BED);
    fadeToBlack();
    glbAtlas()->resetAll();
    intrinsic().clearAll();
    moveTo(p);
    // Reset our orientation on death.
    setFacing(DIR::north());
    readBook(text_lookup("death", death));
    fadeFromBlack();
}

bool
MOB::actionAttack(DIR dir)
{
    POS		newpos = pos().delta(dir);
    MOB		*mob = newpos.mob();

    if (!mob)
	return false;

    if (isAvatar())
    {
	// Avatar is not much of a fighter!
	if (mob->defn().ai == AI_BEEZLEBUB)
	{
	    if (mob->intrinsic().get(INTRINSIC_RIDDLESOLVED))
	    {
		mob->actionYell("You may pass.");
	    }
	    else
	    {
		mob->actionYell("What is round like an apple,\ndeep like a cup,\nand all the King's horses\ncan't pull it up?");
	    }
	    return true;
	}
	else
	{
	    actionEmote("*misses*");
	    return true;
	}
    }

    {
	BUF		attack;
	attack.sprintf("*%s*", defn().attack);
	actionEmote(attack.buffer());
	mob->die(defn().attack);
    }

    return true;
}

bool
MOB::actionWalk(DIR dir)
{
    POS		newpos = pos().delta(dir);

    if (!canMoveTo(newpos))
	return false;

    myLastWalkDir = dir;

    // Actually move first so we are shown on it.
    moveTo(newpos);

    // See if there is anything special here!
    switch (newpos.item())
    {
	case ITEM_BED:
	    if (isAvatar())
	    {
		actionEmote("*zzz*");
		fadeToBlack();
		readBook(text_lookup("item", glb_itemdefs[newpos.item()].name));
		fadeFromBlack();
	    }
	    break;
	case ITEM_CORPSE:
	    if (isAvatar())
	    {
		actionEmote("*searches*");
		readBook(text_lookup("item", glb_itemdefs[newpos.item()].name));
	    }
	    break;

	case ITEM_BONES:
	    if (isAvatar())
	    {
		actionEmote("*searches*");
		readBook(text_lookup("item", glb_itemdefs[newpos.item()].name));
	    }
	    break;

	case ITEM_DREAMEND:
	    if (isAvatar())
	    {
		readBook(text_lookup("game", "victory"));
		fadeToBlack();
		gfx_shutdown();
		exit(0);
	    }
	    break;
	case ITEM_BOULDERSIGN:
	    if (isAvatar())
	    {
		BUF		post;
		PTRLIST<int>	links;
		pos().getLinks(links);

		if (!links.entries())
		{
		    post.reference("blank");
		}
		else
		{
		    post.sprintf("%d", links(0));
		    if (!text_hasentry("sign", post))
			post.reference("foreign");
		}

		actionEmote("*reads*");
		readBook(text_lookup("sign", post));
	    }
	    break;

	case ITEM_HUT:
	    if (isAvatar())
	    {
		// Determine our linked area..

		PTRLIST<int>	links;
		pos().getLinks(links);

		if (!links.entries() || (links(0) <= 0 || links(0) >= NUM_POIS))
		{
		    actionEmote("*empty*");
		}
		else
		{
		    // Find which map it is.
		    POI_NAMES	poi = (POI_NAMES) links(0);

		    MAP		*map = glbAtlas()->findMap(glb_poidefs[poi].name);

		    if (!map)
		    {
			actionEmote("*closed*");
		    }
		    else
		    {
			// LINK101 is reserved for our entrance.
			POSLIST		poslist;
			map->findLinks(101, poslist);

			if (!poslist.entries())
			{
			    actionEmote("*blocked*");
			}
			else
			{
			    newpos = poslist(0);
			    actionEmote("*enters*");
			}
		    }
		}
	    }
	    break;

	case ITEM_STEPPING_DISK:
	{
	    actionEmote("*yanked into space*");
	    myDiskSourceMap = pos().map();
	    MAP		*map = glbAtlas()->findMap("silvery");
	    newpos = map->findRandomSquare(SQUARE_STARS);
	    break;
	}
	case ITEM_LADDERUP:
	{
	    POS		p = newpos.up();
	    if (p.valid())
	    {
		actionEmote("*climbs*");
		newpos = p;
	    }
	    else
		actionEmote("*ladder is blocked*");
	    break;
	}
	case ITEM_LADDERDOWN:
	{
	    POS		p = newpos.down();
	    if (p.valid())
	    {
		actionEmote("*climbs*");
		newpos = p;
	    }
	    else
		actionEmote("*ladder is blocked*");
	    break;
	}

	case ITEM_BOOK_MALACHITETOWER:
	    if (isAvatar())
	    {
		actionEmote("*reads*");
		readBook(text_lookup("item", glb_itemdefs[newpos.item()].name));
	    }
	    break;
    }

    if (glb_itemdefs[newpos.item()].spell != SPELL_NONE)
    {
	actionEmote("*reads*");
	readBook(text_lookup("spell", glb_spelldefs[glb_itemdefs[newpos.item()].spell].name));
    }
    if (glb_itemdefs[newpos.item()].poi != POI_NONE)
    {
	actionEmote("*reads*");
	readBook(text_lookup("poi", glb_poidefs[glb_itemdefs[newpos.item()].poi].name));
    }

    switch (newpos.tile())
    {
	case SQUARE_ENTRANCEDOOR:
	{
	    actionEmote("*leaves*");
	    MAP		*map = glbAtlas()->findMap("world");
	    POI_NAMES	poi;
	    FOREACH_POI(poi)
	    {
		if (!strcmp(glb_poidefs[poi].name, pos().map()->defn().name))
		{
		    // Found it!
		    break;
		}
	    }
	    if (poi < NUM_POIS)
	    {
		POSLIST		poslist;
		map->findLinks(poi, poslist);
		if (!poslist.entries())
		{
		    actionEmote("*locked*");
		}
		else
		{
		    newpos = poslist(0);
		    actionEmote("*leaves*");
		}
	    }
	    else
	    {
		actionEmote("*blocked*");
	    }
	    break;
	}
	case SQUARE_FLOOR_TELE:
	case SQUARE_GROUND_TELE:
	{
	    actionEmote("*teleports*");
	    POSLIST		destlist;
	    newpos.findLinkedSquares(destlist);
	    for (int i = 0; i < destlist.entries(); i++)
	    {
		if (destlist(i).x() != newpos.x() ||
		    destlist(i).y() != newpos.y())
		{
		    newpos = destlist(i);
		    break;
		}
	    }
	    break;
	}
	case SQUARE_FLOOR_SPINNER:
	case SQUARE_GROUND_SPINNER:
	case SQUARE_FOREST_SPINNER:
	{
	    actionEmote("*dizzy*");
	    setFacing(DIR::rand());
	    break;
	}
	case SQUARE_FLOOR_HOLE_TRAP:
	case SQUARE_GROUND_HOLE_TRAP:
	case SQUARE_FLOOR_HOLE:
	case SQUARE_GROUND_HOLE:
	{
	    POS		p = newpos.down();
	    if (p.valid())
	    {
		actionEmote("*falls*");
		newpos = p;
		// Check to see if we have feather fall
		if (intrinsic().get(INTRINSIC_FEATHERFALL))
		{
		    intrinsic().clear(INTRINSIC_FEATHERFALL);
		}
		else
		{
		    die("fall");
		    return true;
		}
	    }
	    else
		actionEmote("*hole is blocked*");
	    break;
	}
	case SQUARE_WATER:
	{
	    if (!intrinsic().get(INTRINSIC_WATERWALK))
	    {
		actionEmote("*drowns*");
		die("drown");
		return true;
	    }
	    intrinsic().clear(INTRINSIC_WATERWALK);
	    break;
	}
	case SQUARE_LAVA:
	{
	    if (!intrinsic().get(INTRINSIC_RESISTFIRE))
	    {
		actionEmote("*burns*");
		die("burn");
		return true;
	    }
	    intrinsic().clear(INTRINSIC_RESISTFIRE);
	    if (!intrinsic().get(INTRINSIC_WATERWALK))
	    {
		actionEmote("*drowns*");
		die("drown");
		return true;
	    }
	    intrinsic().clear(INTRINSIC_WATERWALK);
	    break;
	}
	case SQUARE_STARS:
	case SQUARE_STARS2:
	case SQUARE_STARS3:
	case SQUARE_STARS4:
	case SQUARE_STARS5:
	{
	    if (rand_chance(5) && myDiskSourceMap)
	    {
		actionEmote("*yanked back*");
		newpos = myDiskSourceMap->findRandomItem(ITEM_STEPPING_DISK);
	    }
	    break;
	}
    }
    // Move again in case of a double move
    moveTo(newpos);
    return true;
}

bool
MOB::actionYell(const char *msg)
{
    if (defn().speaker != MOB_NONE &&
	defn().speaker != getDefinition())
    {
	MOB	*speaker = pos().map()->findMobByType((MOB_NAMES)defn().speaker);
	if (speaker)
	    return speaker->actionYell(msg);
    }

    BUF		yell;
    yell.strcpy(msg);
    yell.stripWS();

    if (!yell.isstring())
    {
	// Empty yell
	return false;
    }
    myMessage.strcpy(yell);

    // Check if we are next to the big B.
    if (isAvatar())
    {
	if (pos().east().mob())
	{
	    MOB		*mob = pos().east().mob();
	    if (mob->defn().ai == AI_BEEZLEBUB)
	    {
		// Check the code word!
		BUF	 well, awell;
		well.reference("well");
		awell.reference("a well");
		if (!well.strcasecmp(msg) || !awell.strcasecmp(msg))
		{
		    mob->actionEmote("*growls*");
		    mob->actionYell("BACK TO GOOGLE WITH YOU!");
		    readBook(text_lookup("meaning", "well"));
		}
		else
		{
		    if (!RUNES::getRunes()->riddleRune().strcasecmp(msg))
		    {
			mob->actionYell("You have answered the riddle!\nYou may pass!");

			// This is ugly as hell, but I assure you it
			// beats proper support for 2x2 monsters!
			MOB	*ur = pos().map()->findMobByType(MOB_BAEZLBUB_UR);
			ur->intrinsic().set(INTRINSIC_RIDDLESOLVED);
			ur->moveTo(ur->pos().north().north());

			ur = pos().map()->findMobByType(MOB_BAEZLBUB_UL);
			ur->intrinsic().set(INTRINSIC_RIDDLESOLVED);
			ur->moveTo(ur->pos().north().north());

			ur = pos().map()->findMobByType(MOB_BAEZLBUB_LL);
			ur->intrinsic().set(INTRINSIC_RIDDLESOLVED);
			ur->moveTo(ur->pos().north().north());

			ur = pos().map()->findMobByType(MOB_BAEZLBUB_LR);
			ur->intrinsic().set(INTRINSIC_RIDDLESOLVED);
			ur->moveTo(ur->pos().north().north());
		    }
		    else
		    {
			mob->actionYell("That is not the answer I seek.");
		    }
		}
	    }
	}
    }
    switch (pos().tile())
    {
	case SQUARE_STARS:
	case SQUARE_STARS2:
	case SQUARE_STARS3:
	case SQUARE_STARS4:
	case SQUARE_STARS5:
	{
	    // See if we can jump back!
	    MAP		*map = 
#ifdef STILL_TESTING
		glbAtlas()->findMapByQuest(yell.buffer());
#else
		0;
#endif

	    if (!map)
	    {
		POI_NAMES	poi;
		poi = RUNES::getRunes()->findDisk(yell);
		if (poi != POI_NONE)
		{
		    map = glbAtlas()->findMapByQuest(glb_poidefs[poi].name);
		}
	    }

	    if (map)
	    {
		POS	p = map->findRandomItem(ITEM_STEPPING_DISK);
		if (p.valid())
		{
		    actionEmote("*teleports*");
		    moveTo(p);
		}
	    }
	}
    }

    // Check for magic words
#ifdef STILL_TESTING
    {
	SPELL_NAMES	spell;
	FOREACH_SPELL(spell)
	{
	    if (spell == SPELL_NONE)
		continue;

	    if (!yell.strcasecmp(glb_spelldefs[spell].words))
	    {
		actionCast(spell);
	    }
	}
    }
#endif
    {
	SPELL_NAMES	spell;

	spell = RUNES::getRunes()->findSpell(yell);
	if (spell != SPELL_NONE)
	{
	    actionCast(spell);
	}
    }
    
    return true;
}

bool
MOB::actionCast(SPELL_NAMES spell)
{
    switch (spell)
    {
	case NUM_SPELLS:
	case SPELL_NONE:
	    break;

	case SPELL_RECALL:
	{
	    MAP *map = glbAtlas()->findMap("home");
	    POS p = map->findRandomItem(ITEM_BED);
	    actionEmote(glb_spelldefs[spell].emote);
	    moveTo(p);
	    // It is very mean to leave people disoriented!
	    setFacing(DIR::north());
	    break;
	}
	case SPELL_REVEAL:
	{
	    actionEmote(glb_spelldefs[spell].emote);
	    intrinsic().setTimed(INTRINSIC_REVEAL, 6);
	    break;
	}
	case SPELL_INVIS:
	{
	    if (!intrinsic().get(INTRINSIC_INVIS))
	    {
		actionEmote(glb_spelldefs[spell].emote);
		intrinsic().setTimed(INTRINSIC_INVIS, 20);
	    }
	    break;
	}
	case SPELL_FACENORTH:
	{
	    actionEmote(glb_spelldefs[spell].emote);
	    setFacing(DIR::north());
	    break;
	}
	case SPELL_RESISTFIRE:
	{
	    if (!intrinsic().get(INTRINSIC_RESISTFIRE))
	    {
		actionEmote(glb_spelldefs[spell].emote);
		intrinsic().set(INTRINSIC_RESISTFIRE);
	    }
	    break;
	}
	case SPELL_WATERWALK:
	{
	    if (!intrinsic().get(INTRINSIC_WATERWALK))
	    {
		actionEmote(glb_spelldefs[spell].emote);
		intrinsic().set(INTRINSIC_WATERWALK);
	    }
	    break;
	}
	case SPELL_FEATHERFALL:
	{
	    if (!intrinsic().get(INTRINSIC_FEATHERFALL))
	    {
		actionEmote(glb_spelldefs[spell].emote);
		intrinsic().set(INTRINSIC_FEATHERFALL);
	    }
	    break;
	}

	case SPELL_BIRDSVIEW:
	{
	    actionEmote(glb_spelldefs[spell].emote);

	    // What did I say about never using map()?
	    // I guess I lied :>
	    if (pos().map())
	    {
		u8		*rgb;
		int		 handle;

		rgb = pos().map()->buildMiniMap();
		handle = gfx_overlay_create(rgb, 
				pos().map()->width(), pos().map()->height(),
				0, 0, 4, 
				GFX_OVERLAY_TOP, false);
		delete [] rgb;
		gfx_overlay_center(handle);

		// Block for a key
		redrawWorld();
		gfx_getKey(true);

		gfx_overlay_destroy(handle);
	    }
		
	    break;
	}
	case SPELL_FIREBALL:
	{
	    actionEmote(glb_spelldefs[spell].emote);

	    POS		firepos = pos().delta(myLastWalkDir);

	    if (firepos.valid())
	    {
		firepos.setFlag(FLAG_FIRE, true);
		firepos.setFlag(FLAG_FIREFRESH, true);
	    }

	    break;
	}

    }
    return false;
}

bool
MOB::actionEmote(const char *msg)
{
    if (defn().speaker != MOB_NONE &&
	defn().speaker != getDefinition())
    {
	MOB	*speaker = pos().map()->findMobByType((MOB_NAMES)defn().speaker);
	if (speaker)
	    return speaker->actionEmote(msg);
    }
    myEmote.strcpy(msg);

    return true;
}

extern MOB *glbAvatar;

bool
MOB::heartbeat(int phase)
{
    // Avoid double moves caused by crossing map boundaries.
    if (phase == myLastPhase)
	return false;

    myLastPhase = phase;

    // Check for burning
    if (pos().getFlag(FLAG_FIRE) ||
	pos().defn().burning)
    {
	if (intrinsic().get(INTRINSIC_RESISTFIRE))
	    actionEmote("*resists*");
	else
	{
	    actionEmote("*burns*");
	    if (isAvatar())
	    {
		die("burn");
		return false;
	    }

	    if (pos().item() == ITEM_NONE)
		pos().setItem(ITEM_CORPSE);

	    moveTo(POS::invalidPos());
	    delete this;
	    return false;
	}
    }

    intrinsic().heartbeat();

    return true;
}

bool
MOB::doAI()
{
    if (isAvatar())
	return false;

    if (defn().ai != AI_CHARGE)
	return false;

    if (glbAvatar && !glbAvatar->intrinsic().get(INTRINSIC_INVIS))
    {
	DISTMAP		dist(glbAvatar->pos());

	DIR		d;
	if (dist.grad(pos(), d))
	{
	    // Check if we are beside!
	    if (pos().delta(d).mob() == glbAvatar)
	    {
		return actionAttack(d);
	    }
	    if (actionWalk(d))
		return true;
	}
    }

    return aiWander();
}

bool
MOB::aiWander()
{
    DIR		dir, ndir;
    int		valid = 0;

    FORALL_DIR(dir)
    {
	if (canMoveIn(dir))
	{
	    if (!rand_choice(++valid))
		ndir = dir;
	}
    }

    if (valid)
    {
	return actionWalk(ndir);
    }

    return false;
}

bool
MOB::canMoveTo(POS newpos) const
{
    if (newpos.mob())
	return false;

    if (!newpos.defn().passable)
	return false;

    // Monsters are smart enough to not suicide.
    if (!isAvatar())
    {
	if (newpos.defn().dangerous)
	    return false;

	if (newpos.getFlag(FLAG_FIRE))
	    return false;
    }

    return true;
}
